#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#***************************************************************************
#                                  _   _ ____  _
#  Project                     ___| | | |  _ \| |
#                             / __| | | | |_) | |
#                            | (__| |_| |  _ <| |___
#                             \___|\___/|_| \_\_____|
#
# Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://curl.se/docs/copyright.html.
#
# You may opt to use, copy, modify, merge, publish, distribute and/or sell
# copies of the Software, and permit persons to whom the Software is
# furnished to do so, under the terms of the COPYING file.
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
# KIND, either express or implied.
#
# SPDX-License-Identifier: curl
#
###########################################################################
#
import logging
import pytest

from testenv import Env, CurlClient


log = logging.getLogger(__name__)


@pytest.mark.skipif(condition=not Env.httpd_is_at_least('2.4.55'),
                    reason=f"httpd version too old for this: {Env.httpd_version()}")
class TestErrors:

    # download 1 file, check that we get CURLE_PARTIAL_FILE
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_05_01_partial_1(self, env: Env, httpd, nghttpx, proto):
        if proto == 'h3' and not env.have_h3():
            pytest.skip("h3 not supported")
        count = 1
        curl = CurlClient(env=env)
        urln = f'https://{env.authority_for(env.domain1, proto)}' \
            f'/curltest/tweak?id=[0-{count - 1}]'\
            '&chunks=3&chunk_size=16000&body_error=reset'
        r = curl.http_download(urls=[urln], alpn_proto=proto, extra_args=[
            '--retry', '0'
        ])
        r.check_exit_code(False)
        invalid_stats = []
        for idx, s in enumerate(r.stats):
            if 'exitcode' not in s or s['exitcode'] not in [18, 56, 92, 95]:
                invalid_stats.append(f'request {idx} exit with {s["exitcode"]}')
        assert len(invalid_stats) == 0, f'failed: {invalid_stats}'

    # download files, check that we get CURLE_PARTIAL_FILE for all
    @pytest.mark.parametrize("proto", ['h2', 'h3'])
    def test_05_02_partial_20(self, env: Env, httpd, nghttpx, proto):
        if proto == 'h3' and not env.have_h3():
            pytest.skip("h3 not supported")
        if proto == 'h3' and env.curl_uses_ossl_quic():
            pytest.skip("openssl-quic is flaky in yielding proper error codes")
        count = 20
        curl = CurlClient(env=env)
        urln = f'https://{env.authority_for(env.domain1, proto)}' \
            f'/curltest/tweak?id=[0-{count - 1}]'\
            '&chunks=5&chunk_size=16000&body_error=reset'
        r = curl.http_download(urls=[urln], alpn_proto=proto, extra_args=[
            '--retry', '0', '--parallel',
        ])
        r.check_exit_code(False)
        assert len(r.stats) == count, f'did not get all stats: {r}'
        invalid_stats = []
        for idx, s in enumerate(r.stats):
            if 'exitcode' not in s or s['exitcode'] not in [18, 55, 56, 92, 95]:
                invalid_stats.append(f'request {idx} exit with {s["exitcode"]}\n{s}')
        assert len(invalid_stats) == 0, f'failed: {invalid_stats}'

    # access a resource that, on h2, RST the stream with HTTP_1_1_REQUIRED
    def test_05_03_required(self, env: Env, httpd, nghttpx):
        curl = CurlClient(env=env)
        proto = 'http/1.1'
        urln = f'https://{env.authority_for(env.domain1, proto)}/curltest/1_1'
        r = curl.http_download(urls=[urln], alpn_proto=proto)
        r.check_exit_code(0)
        r.check_response(http_status=200, count=1)
        proto = 'h2'
        urln = f'https://{env.authority_for(env.domain1, proto)}/curltest/1_1'
        r = curl.http_download(urls=[urln], alpn_proto=proto)
        r.check_exit_code(0)
        r.check_response(http_status=200, count=1)
        # check that we did a downgrade
        assert r.stats[0]['http_version'] == '1.1', r.dump_logs()

    # On the URL used here, Apache is doing an "unclean" TLS shutdown,
    # meaning it sends no shutdown notice and just closes TCP.
    # The HTTP response delivers a body without Content-Length. We expect:
    # - http/1.0 to fail since it relies on a clean connection close to
    #   detect the end of the body
    # - http/1.1 to work since it will used "chunked" transfer encoding
    #   and stop receiving when that signals the end
    # - h2 to work since it will signal the end of the response before
    #   and not see the "unclean" close either
    @pytest.mark.parametrize("proto", ['http/1.0', 'http/1.1', 'h2'])
    def test_05_04_unclean_tls_shutdown(self, env: Env, httpd, nghttpx, proto):
        if proto == 'h3' and not env.have_h3():
            pytest.skip("h3 not supported")
        count = 10 if proto == 'h2' else 1
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}'\
            f'/curltest/shutdown_unclean?id=[0-{count-1}]&chunks=4'
        r = curl.http_download(urls=[url], alpn_proto=proto, extra_args=[
            '--parallel', '--trace-config', 'ssl'
        ])
        if proto == 'http/1.0':
            # we are inconsistent if we fail or not in missing TLS shutdown
            # openssl code ignore such errors intentionally in non-debug builds
            r.check_exit_code(56)
        else:
            r.check_exit_code(0)
            r.check_response(http_status=200, count=count)



#
# dotex: "I don't care, just do the tex or whatever..."
#
# Written in 2025 by Lexi Mayfield
# Released into the Public Domain, see UNLICENSE.txt
#

import io
import logging
import struct
from array import array
from pathlib import Path

import dotex.png as png
from dotex.palette import DoomPalette


class DoomPatch:
    name: str
    width: int
    height: int
    left: int
    top: int
    data: array  # Array of uint containing RGBA data.

    def __init__(self):
        self.name = ""
        self.width = 0
        self.height = 0
        self.left = 0
        self.top = 0
        self.data = array("L")

    def __repr__(self) -> str:
        return f"<DoomGraphic name={self.name} res={self.width}x{self.height} left={self.left} top={self.top}>"

    def set_left_top(self, left: int, top: int) -> None:
        """
        Set the left and top attributes of the graphic.  You should get these
        as a pair.
        """

        self.left = left
        self.top = top

    def read_image(self, file: Path) -> None:
        """
        Given a path, read in the image data.
        """
        logging.debug(f"Reading '{file}' for patch.")

        reader = png.Reader(filename=file)
        self.name = file.stem
        self.width, self.height, pixels, metadata = reader.read()
        if "palette" in metadata:
            # Paletted image
            for row in pixels:
                for pixel in row:
                    color = metadata["palette"][pixel]
                    co = color[0]
                    co |= color[1] << 8
                    co |= color[2] << 16
                    if len(color) == 4:
                        co |= color[3] << 24
                    else:
                        co |= 0xFF << 24
                    self.data.append(co)
        elif metadata["greyscale"] is True:
            # Greyscale image
            for row in pixels:
                for pixel in row:
                    co = pixel
                    co |= pixel << 8
                    co |= pixel << 16
                    co |= 0xFF << 24
                    self.data.append(co)
        else:
            raise Exception("Non-palette PNG not supported (yet)")

    def to_bytes(self, pal: DoomPalette) -> bytes:
        """
        Convert raw graphic data to Doom Patch format.

        Based on the pseudo-code from https://doomwiki.org/wiki/Picture_format
        """
        logging.debug(f"Writing patch '{self.name}'.")

        data = io.BytesIO()

        # Patch header
        data.write(struct.pack("<HHhh", self.width, self.height, self.left, self.top))

        # Keep track of column offsets.
        columnofs: list[int] = []
        columns_off = data.tell()

        # Write out an empty columnofs for now.
        data.write(b"\x00\x00\x00\x00" * self.width)
        for x in range(self.width):
            # Align start of column data to four bytes
            while data.tell() % 4 != 0:
                data.write(b"\x00")

            # Note the start of the current column.
            columnofs.append(data.tell())

            y = 0
            post_start = 0
            post_len = 0
            last_icolor: int | None = None

            def end_post():
                """
                End the current post.
                """
                assert last_icolor is not None
                data.write(struct.pack("<B", last_icolor))
                saved = data.tell()
                data.seek(post_start + 1)
                data.write(struct.pack("<B", post_len))
                data.seek(saved)

            while y < self.height:
                # Break the pixel into RGB and A.
                rgba = self.data[y * self.width + x]
                rgb = rgba & 0xFFFFFF
                a = (rgba & 0xFF000000) >> 24

                # Figure out our color by index.
                icolor: int | None
                if a > 0:
                    icolor = pal.playpal.get(rgb)
                    if icolor is None:
                        icolor = pal.add_close_color(rgb)
                else:
                    icolor = None

                if icolor is not None and post_len == 0:
                    # Got pixel data, but we need to start the post header.
                    post_start = data.tell()
                    data.write(struct.pack("<BBBB", y, 0xFF, icolor, icolor))
                    last_icolor = icolor
                    post_len += 1
                elif icolor is not None and post_len > 0:
                    # In the middle of the post.
                    data.write(struct.pack("<B", icolor))
                    last_icolor = icolor
                    post_len += 1
                elif icolor is None and post_len > 0:
                    # Write the end of a post.
                    end_post()
                    post_len = 0

                # Advance to the next pixel
                y += 1

            if post_len > 0:
                # Write the end of a post.
                end_post()

            # Write end-of-column flag footer.
            data.write(b"\xff")

        # Write the column offsets.
        data.seek(columns_off)
        for col in columnofs:
            data.write(struct.pack("<I", col))

        data.seek(0)
        return data.read()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#***************************************************************************
#                                  _   _ ____  _
#  Project                     ___| | | |  _ \| |
#                             / __| | | | |_) | |
#                            | (__| |_| |  _ <| |___
#                             \___|\___/|_| \_\_____|
#
# Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://curl.se/docs/copyright.html.
#
# You may opt to use, copy, modify, merge, publish, distribute and/or sell
# copies of the Software, and permit persons to whom the Software is
# furnished to do so, under the terms of the COPYING file.
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
# KIND, either express or implied.
#
# SPDX-License-Identifier: curl
#
###########################################################################
#
import logging
import os
import socket
import subprocess
import time
from datetime import timedelta, datetime
from json import JSONEncoder
from typing import Dict

from .curl import CurlClient
from .env import Env
from .ports import alloc_ports_and_do

log = logging.getLogger(__name__)


class Caddy:

    PORT_SPECS = {
        'caddy': socket.SOCK_STREAM,
        'caddys': socket.SOCK_STREAM,
    }

    def __init__(self, env: Env):
        self.env = env
        self._caddy = os.environ['CADDY'] if 'CADDY' in os.environ else env.caddy
        self._caddy_dir = os.path.join(env.gen_dir, 'caddy')
        self._docs_dir = os.path.join(self._caddy_dir, 'docs')
        self._conf_file = os.path.join(self._caddy_dir, 'Caddyfile')
        self._error_log = os.path.join(self._caddy_dir, 'caddy.log')
        self._tmp_dir = os.path.join(self._caddy_dir, 'tmp')
        self._process = None
        self._http_port = 0
        self._https_port = 0
        self._rmf(self._error_log)

    @property
    def docs_dir(self):
        return self._docs_dir

    @property
    def port(self) -> int:
        return self._https_port

    def clear_logs(self):
        self._rmf(self._error_log)

    def is_running(self):
        if self._process:
            self._process.poll()
            return self._process.returncode is None
        return False

    def start_if_needed(self):
        if not self.is_running():
            return self.start()
        return True

    def initial_start(self):

        def startup(ports: Dict[str, int]) -> bool:
            self._http_port = ports['caddy']
            self._https_port = ports['caddys']
            if self.start():
                self.env.update_ports(ports)
                return True
            self.stop()
            self._http_port = 0
            self._https_port = 0
            return False

        return alloc_ports_and_do(Caddy.PORT_SPECS, startup,
                                  self.env.gen_root, max_tries=3)

    def start(self, wait_live=True):
        assert self._http_port > 0 and self._https_port > 0
        self._mkpath(self._tmp_dir)
        if self._process:
            self.stop()
        self._write_config()
        args = [
            self._caddy, 'run'
        ]
        caddyerr = open(self._error_log, 'a')
        self._process = subprocess.Popen(args=args, cwd=self._caddy_dir, stderr=caddyerr)
        if self._process.returncode is not None:
            return False
        return not wait_live or self.wait_live(timeout=timedelta(seconds=Env.SERVER_TIMEOUT))

    def stop(self, wait_dead=True):
        self._mkpath(self._tmp_dir)
        if self._process:
            self._process.terminate()
            self._process.wait(timeout=2)
            self._process = None
            return not wait_dead or self.wait_dead(timeout=timedelta(seconds=5))
        return True

    def restart(self):
        self.stop()
        return self.start()

    def wait_dead(self, timeout: timedelta):
        curl = CurlClient(env=self.env, run_dir=self._tmp_dir)
        try_until = datetime.now() + timeout
        while datetime.now() < try_until:
            check_url = f'https://{self.env.domain1}:{self.port}/'
            r = curl.http_get(url=check_url)
            if r.exit_code != 0:
                return True
            log.debug(f'waiting for caddy to stop responding: {r}')
            time.sleep(.1)
        log.debug(f"Server still responding after {timeout}")
        return False

    def wait_live(self, timeout: timedelta):
        curl = CurlClient(env=self.env, run_dir=self._tmp_dir)
        try_until = datetime.now() + timeout
        while datetime.now() < try_until:
            check_url = f'https://{self.env.domain1}:{self.port}/'
            r = curl.http_get(url=check_url)
            if r.exit_code == 0:
                return True
            time.sleep(.1)
        log.error(f"Caddy still not responding after {timeout}")
        return False

    def _rmf(self, path):
        if os.path.exists(path):
            return os.remove(path)

    def _mkpath(self, path):
        if not os.path.exists(path):
            return os.makedirs(path)

    def _write_config(self):
        domain1 = self.env.domain1
        creds1 = self.env.get_credentials(domain1)
        assert creds1  # convince pytype this isn't None
        domain2 = self.env.domain2
        creds2 = self.env.get_credentials(domain2)
        assert creds2  # convince pytype this isn't None
        self._mkpath(self._docs_dir)
        self._mkpath(self._tmp_dir)
        with open(os.path.join(self._docs_dir, 'data.json'), 'w') as fd:
            data = {
                'server': f'{domain1}',
            }
            fd.write(JSONEncoder().encode(data))
        with open(self._conf_file, 'w') as fd:
            conf = [   # base server config
                '{',
                f'  http_port {self._http_port}',
                f'  https_port {self._https_port}',
                '  log default {',
                '     level ERROR',
                '}',
                f'  servers :{self._https_port} {{',
                '    protocols h3 h2 h1',
                '  }',
                '}',
                f'{domain1}:{self._https_port} {{',
                '  file_server * {',
                f'    root {self._docs_dir}',
                '  }',
                f'  tls {creds1.cert_file} {creds1.pkey_file}',
                '}',
            ]
            if self.env.http_port > 0:
                conf.extend([
                    f'{domain2} {{',
                    f'  reverse_proxy /* http://localhost:{self.env.http_port} {{',
                    '  }',
                    f'  tls {creds2.cert_file} {creds2.pkey_file}',
                    '}',
                ])
            fd.write("\n".join(conf))



#
# dotex: "I don't care, just do the tex or whatever..."
#
# Written in 2025 by Lexi Mayfield
# Released into the Public Domain, see UNLICENSE.txt
#

import io
import logging
from pathlib import Path
from typing import Dict

from dotex.util import rgb_pack, rgb_unpack


class DoomPalette:
    playpal: Dict[int, int]

    def __init__(self):
        self.playpal = {}

    def _read_gimp_colors(self, fh: io.TextIOBase) -> None:
        """
        Read the main GIMP color palette colors.
        """
        index = 0
        while True:
            line = fh.readline()
            if line == "":
                break
            if line == "\n":
                continue
            elif line.startswith("#"):
                continue

            color = line.split()
            if len(color) < 3:
                raise Exception(f"Line '{line}' does not contain a color")

            rgb = rgb_pack(int(color[0]), int(color[1]), int(color[2]))
            self.playpal[rgb] = index
            index += 1

    def read_gimp_palette(self, file: Path) -> None:
        """
        Given a path, read a palette in GIMP palette format.

        See: https://developer.gimp.org/core/standards/gpl/
        """
        logging.debug(f"Reading '{file}' for palette.")

        with open(file, "rt") as fh:
            line = fh.readline().rstrip()
            if line != "GIMP Palette":
                raise Exception(f"{file.name} does not contain GIMP Palette")

            saved = fh.tell()
            line = fh.readline().rstrip("\r")
            if not line.startswith("Name:"):
                fh.seek(saved)
                return self._read_gimp_colors(fh)

            saved = fh.tell()
            line = fh.readline().rstrip("\r")
            if not line.startswith("Columns:"):
                fh.seek(saved)
                return self._read_gimp_colors(fh)

            return self._read_gimp_colors(fh)

    def add_close_color(self, rgb: int) -> int:
        """
        Seek out the closest color to a given palette index.
        """
        logging.debug(f"Finding color close to {rgb:06x}")

        r, g, b = rgb_unpack(rgb)

        closest_index: int | None = None
        closest_distsq: float = float("inf")
        for color in self.playpal.keys():
            # [LM] Calling rgb_unpack slows this down
            cr = color & 0xFF
            cg = (color >> 8) & 0xFF
            cb = (color >> 16) & 0xFF

            dr = r - cr
            dg = g - cg
            db = b - cb

            distsq = dr * dr + dg * dg + db * db
            if distsq < closest_distsq:
                closest_index = self.playpal[color]
                closest_distsq = distsq

        assert closest_index is not None
        self.playpal[rgb] = closest_index
        return closest_index



# This import depends on the automake rule protoc_middleman, please make sure
# protoc_middleman has been built before run this file.
import argparse
import json
import re
import os.path
# BEGIN OPENSOURCE
import sys
sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
# END OPENSOURCE
import tmp.benchmarks_pb2 as benchmarks_pb2

__file_size_map = {}

def __get_data_size(filename):
  if filename[0] != '/':
    filename = os.path.dirname(os.path.abspath(__file__)) + "/../" + filename
  if filename in __file_size_map:
    return __file_size_map[filename]
  benchmark_dataset = benchmarks_pb2.BenchmarkDataset()
  benchmark_dataset.ParseFromString(
      open(filename, "rb").read())
  size = 0
  count = 0
  for payload in benchmark_dataset.payload:
    size += len(payload)
    count += 1
  __file_size_map[filename] = (size, 1.0 * size / count)
  return size, 1.0 * size / count


def __extract_file_name(file_name):
  name_list = re.split(r"[/\.]", file_name)
  short_file_name = ""
  for name in name_list:
    if name[:14] == "google_message":
      short_file_name = name
  return short_file_name


__results = []


# CPP results example:
# [
#   "benchmarks": [
#     {
#       "bytes_per_second": int,
#       "cpu_time_ns": double,
#       "iterations": int,
#       "name: string,
#       "real_time_ns: double,
#       ...
#     },
#     ...
#   ],
#   ...
# ]
def __parse_cpp_result(filename):
  if filename == "":
    return
  if filename[0] != '/':
    filename = os.path.dirname(os.path.abspath(__file__)) + '/' + filename
  with open(filename, encoding="utf-8") as f:
    results = json.loads(f.read())
    for benchmark in results["benchmarks"]:
      data_filename = "".join(
          re.split("(_parse_|_serialize)", benchmark["name"])[0])
      behavior = benchmark["name"][len(data_filename) + 1:]
      if data_filename[:2] == "BM":
        data_filename = data_filename[3:]
      __results.append({
        "language": "cpp",
        "dataFilename": data_filename,
        "behavior": behavior,
        "throughput": benchmark["bytes_per_second"] / 2.0 ** 20
      })


# Synthetic benchmark results example:
# [
#   "benchmarks": [
#     {
#       "cpu_time_ns": double,
#       "iterations": int,
#       "name: string,
#       "real_time_ns: double,
#       ...
#     },
#     ...
#   ],
#   ...
# ]
def __parse_synthetic_result(filename):
  if filename == "":
    return
  if filename[0] != "/":
    filename = os.path.dirname(os.path.abspath(__file__)) + "/" + filename
  with open(filename, encoding="utf-8") as f:
    results = json.loads(f.read())
    for benchmark in results["benchmarks"]:
      __results.append({
          "language": "cpp",
          "dataFilename": "",
          "behavior": "synthetic",
          "throughput": 10.0**9 / benchmark["cpu_time_ns"]
      })


# Python results example:
# [
#   [
#     {
#       "filename": string,
#       "benchmarks": {
#         behavior: results,
#         ...
#       },
#     },
#     ...
#   ], #pure-python
#   ...
# ]
def __parse_python_result(filename):
  if filename == "":
    return
  if filename[0] != '/':
    filename = os.path.dirname(os.path.abspath(__file__)) + '/' + filename
  with open(filename, encoding="utf-8") as f:
    results_list = json.loads(f.read())
    for results in results_list:
      for result in results:
        _, avg_size = __get_data_size(result["filename"])
        for behavior in result["benchmarks"]:
          __results.append({
            "language": "python",
            "dataFilename": __extract_file_name(result["filename"]),
            "behavior": behavior,
            "throughput": result["benchmarks"][behavior]
          })


# Java results example:
# [
#   {
#     "id": string,
#     "instrumentSpec": {...},
#     "measurements": [
#       {
#         "weight": float,
#         "value": {
#           "magnitude": float,
#           "unit": string
#         },
#         ...
#       },
#       ...
#     ],
#     "run": {...},
#     "scenario": {
#       "benchmarkSpec": {
#         "methodName": string,
#         "parameters": {
#            defined parameters in the benchmark: parameters value
#         },
#         ...
#       },
#       ...
#     }
#
#   },
#   ...
# ]
def __parse_java_result(filename):
  if filename == "":
    return
  if filename[0] != '/':
    filename = os.path.dirname(os.path.abspath(__file__)) + '/' + filename
  with open(filename, encoding="utf-8") as f:
    results = json.loads(f.read())
    for result in results:
      total_weight = 0
      total_value = 0
      for measurement in result["measurements"]:
        total_weight += measurement["weight"]
        total_value += measurement["value"]["magnitude"]
      avg_time = total_value * 1.0 / total_weight
      total_size, _ = __get_data_size(
          result["scenario"]["benchmarkSpec"]["parameters"]["dataFile"])
      __results.append({
        "language": "java",
        "throughput": total_size / avg_time * 1e9 / 2 ** 20,
        "behavior": result["scenario"]["benchmarkSpec"]["methodName"],
        "dataFilename": __extract_file_name(
            result["scenario"]["benchmarkSpec"]["parameters"]["dataFile"])
      })


# Go benchmark results:
#
# goos: linux
# goarch: amd64
# Benchmark/.././datasets/google_message2/dataset.google_message2.pb/Unmarshal-12               3000      705784 ns/op
# Benchmark/.././datasets/google_message2/dataset.google_message2.pb/Marshal-12                 2000      634648 ns/op
# Benchmark/.././datasets/google_message2/dataset.google_message2.pb/Size-12                    5000      244174 ns/op
# Benchmark/.././datasets/google_message2/dataset.google_message2.pb/Clone-12                    300     4120954 ns/op
# Benchmark/.././datasets/google_message2/dataset.google_message2.pb/Merge-12                    300     4108632 ns/op
# PASS
# ok    _/usr/local/google/home/yilunchong/mygit/protobuf/benchmarks  124.173s
def __parse_go_result(filename):
  if filename == "":
    return
  if filename[0] != '/':
    filename = os.path.dirname(os.path.abspath(__file__)) + '/' + filename
  with open(filename, encoding="utf-8") as f:
    for line in f:
      result_list = re.split(r"[\ \t]+", line)
      if result_list[0][:9] != "Benchmark":
        continue
      first_slash_index = result_list[0].find('/')
      last_slash_index = result_list[0].rfind('/')
      full_filename = result_list[0][first_slash_index+1:last_slash_index]
      total_bytes, _ = __get_data_size(full_filename)
      behavior_with_suffix = result_list[0][last_slash_index+1:]
      last_dash = behavior_with_suffix.rfind("-")
      if last_dash == -1:
        behavior = behavior_with_suffix
      else:
        behavior = behavior_with_suffix[:last_dash]
      __results.append({
        "dataFilename": __extract_file_name(full_filename),
        "throughput": total_bytes / float(result_list[2]) * 1e9 / 2 ** 20,
        "behavior": behavior,
        "language": "go"
      })


# Self built json results example:
#
# [
#   {
#     "filename": string,
#     "benchmarks": {
#       behavior: results,
#       ...
#     },
#   },
#   ...
# ]
def __parse_custom_result(filename, language):
  if filename == "":
    return
  if filename[0] != '/':
    filename = os.path.dirname(os.path.abspath(__file__)) + '/' + filename
  with open(filename, encoding="utf-8") as f:
    results = json.loads(f.read())
    for result in results:
      _, avg_size = __get_data_size(result["filename"])
      for behavior in result["benchmarks"]:
        __results.append({
          "language": language,
          "dataFilename": __extract_file_name(result["filename"]),
          "behavior": behavior,
          "throughput": result["benchmarks"][behavior]
        })


def __parse_js_result(filename, language):
  return __parse_custom_result(filename, language)

def __parse_php_result(filename, language):
  return __parse_custom_result(filename, language)


def get_result_from_file(cpp_file="",
                         java_file="",
                         python_file="",
                         go_file="",
                         synthetic_file="",
                         node_file="",
                         php_c_file="",
                         php_file=""):
  results = {}
  if cpp_file != "":
    __parse_cpp_result(cpp_file)
  if java_file != "":
    __parse_java_result(java_file)
  if python_file != "":
    __parse_python_result(python_file)
  if go_file != "":
    __parse_go_result(go_file)
  if synthetic_file != "":
    __parse_synthetic_result(synthetic_file)
  if node_file != "":
    __parse_js_result(node_file, "node")
  if php_file != "":
    __parse_php_result(php_file, "php")
  if php_c_file != "":
    __parse_php_result(php_c_file, "php")

  return __results


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument(
      "-cpp",
      "--cpp_input_file",
      help="The CPP benchmark result file's name",
      default="")
  parser.add_argument(
      "-java",
      "--java_input_file",
      help="The Java benchmark result file's name",
      default="")
  parser.add_argument(
      "-python",
      "--python_input_file",
      help="The Python benchmark result file's name",
      default="")
  parser.add_argument(
      "-go",
      "--go_input_file",
      help="The golang benchmark result file's name",
      default="")
  parser.add_argument(
      "-node",
      "--node_input_file",
      help="The node.js benchmark result file's name",
      default="")
  parser.add_argument(
      "-php",
      "--php_input_file",
      help="The pure php benchmark result file's name",
      default="")
  parser.add_argument(
      "-php_c",
      "--php_c_input_file",
      help="The php with c ext benchmark result file's name",
      default="")
  args = parser.parse_args()

  results = get_result_from_file(
      cpp_file=args.cpp_input_file,
      java_file=args.java_input_file,
      python_file=args.python_input_file,
      go_file=args.go_input_file,
      node_file=args.node_input_file,
      php_file=args.php_input_file,
      php_c_file=args.php_c_input_file,
  )
  print(json.dumps(results, indent=2))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#***************************************************************************
#                                  _   _ ____  _
#  Project                     ___| | | |  _ \| |
#                             / __| | | | |_) | |
#                            | (__| |_| |  _ <| |___
#                             \___|\___/|_| \_\_____|
#
# Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://curl.se/docs/copyright.html.
#
# You may opt to use, copy, modify, merge, publish, distribute and/or sell
# copies of the Software, and permit persons to whom the Software is
# furnished to do so, under the terms of the COPYING file.
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
# KIND, either express or implied.
#
# SPDX-License-Identifier: curl
#
###########################################################################
#
import logging
import pytest

from testenv import Env
from testenv import CurlClient


log = logging.getLogger(__name__)


class TestBasic:

    # simple http: GET
    def test_01_01_http_get(self, env: Env, httpd):
        curl = CurlClient(env=env)
        url = f'http://{env.domain1}:{env.http_port}/data.json'
        r = curl.http_get(url=url)
        r.check_response(http_status=200)
        assert r.json['server'] == env.domain1

    # simple https: GET, any http version
    @pytest.mark.skipif(condition=not Env.have_ssl_curl(), reason="curl without SSL")
    def test_01_02_https_get(self, env: Env, httpd):
        curl = CurlClient(env=env)
        url = f'https://{env.domain1}:{env.https_port}/data.json'
        r = curl.http_get(url=url)
        r.check_response(http_status=200)
        assert r.json['server'] == env.domain1

    # simple https: GET, h2 wanted and got
    @pytest.mark.skipif(condition=not Env.have_ssl_curl(), reason="curl without SSL")
    def test_01_03_h2_get(self, env: Env, httpd):
        curl = CurlClient(env=env)
        url = f'https://{env.domain1}:{env.https_port}/data.json'
        r = curl.http_get(url=url, extra_args=['--http2'])
        r.check_response(http_status=200, protocol='HTTP/2')
        assert r.json['server'] == env.domain1

    # simple https: GET, h2 unsupported, fallback to h1
    @pytest.mark.skipif(condition=not Env.have_ssl_curl(), reason="curl without SSL")
    def test_01_04_h2_unsupported(self, env: Env, httpd):
        curl = CurlClient(env=env)
        url = f'https://{env.domain2}:{env.https_port}/data.json'
        r = curl.http_get(url=url, extra_args=['--http2'])
        r.check_response(http_status=200, protocol='HTTP/1.1')
        assert r.json['server'] == env.domain2

    # simple h3: GET, want h3 and get it
    @pytest.mark.skipif(condition=not Env.have_h3(), reason="h3 not supported")
    def test_01_05_h3_get(self, env: Env, httpd, nghttpx):
        curl = CurlClient(env=env)
        url = f'https://{env.domain1}:{env.h3_port}/data.json'
        r = curl.http_get(url=url, extra_args=['--http3-only'])
        r.check_response(http_status=200, protocol='HTTP/3')
        assert r.json['server'] == env.domain1

    # simple download, check connect/handshake timings
    @pytest.mark.skipif(condition=not Env.have_ssl_curl(), reason="curl without SSL")
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_01_06_timings(self, env: Env, httpd, nghttpx, proto):
        if proto == 'h3' and not env.have_h3():
            pytest.skip("h3 not supported")
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}/data.json'
        r = curl.http_download(urls=[url], alpn_proto=proto, with_stats=True)
        r.check_stats(http_status=200, count=1,
                      remote_port=env.port_for(alpn_proto=proto),
                      remote_ip='127.0.0.1')
        # there are cases where time_connect is reported as 0
        assert r.stats[0]['time_connect'] >= 0, f'{r.stats[0]}'
        assert r.stats[0]['time_appconnect'] > 0, f'{r.stats[0]}'

    # simple https: HEAD
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    @pytest.mark.skipif(condition=not Env.have_ssl_curl(), reason="curl without SSL")
    def test_01_07_head(self, env: Env, httpd, nghttpx, proto):
        if proto == 'h3' and not env.have_h3():
            pytest.skip("h3 not supported")
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}/data.json'
        r = curl.http_download(urls=[url], with_stats=True, with_headers=True,
                               extra_args=['-I'])
        r.check_stats(http_status=200, count=1, exitcode=0,
                      remote_port=env.port_for(alpn_proto=proto),
                      remote_ip='127.0.0.1')
        # got the Content-Length: header, but did not download anything
        assert r.responses[0]['header']['content-length'] == '30', f'{r.responses[0]}'
        assert r.stats[0]['size_download'] == 0, f'{r.stats[0]}'

    # http: GET for HTTP/2, see Upgrade:, 101 switch
    def test_01_08_h2_upgrade(self, env: Env, httpd):
        curl = CurlClient(env=env)
        url = f'http://{env.domain1}:{env.http_port}/data.json'
        r = curl.http_get(url=url, extra_args=['--http2'])
        r.check_exit_code(0)
        assert len(r.responses) == 2, f'{r.responses}'
        assert r.responses[0]['status'] == 101, f'{r.responses[0]}'
        assert r.responses[1]['status'] == 200, f'{r.responses[1]}'
        assert r.responses[1]['protocol'] == 'HTTP/2', f'{r.responses[1]}'
        assert r.json['server'] == env.domain1

    # http: GET for HTTP/2 with prior knowledge
    def test_01_09_h2_prior_knowledge(self, env: Env, httpd):
        curl = CurlClient(env=env)
        url = f'http://{env.domain1}:{env.http_port}/data.json'
        r = curl.http_get(url=url, extra_args=['--http2-prior-knowledge'])
        r.check_exit_code(0)
        assert len(r.responses) == 1, f'{r.responses}'
        assert r.response['status'] == 200, f'{r.responsw}'
        assert r.response['protocol'] == 'HTTP/2', f'{r.response}'
        assert r.json['server'] == env.domain1

    # http: strip TE header in HTTP/2 requests
    def test_01_10_te_strip(self, env: Env, httpd):
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, "h2")}/data.json'
        r = curl.http_get(url=url, extra_args=['--http2', '-H', 'TE: gzip'])
        r.check_exit_code(0)
        assert len(r.responses) == 1, f'{r.responses}'
        assert r.responses[0]['status'] == 200, f'{r.responses[1]}'
        assert r.responses[0]['protocol'] == 'HTTP/2', f'{r.responses[1]}'

    # http: large response headers
    # send 48KB+ sized response headers to check we handle that correctly
    # larger than 64KB headers expose a bug in Apache HTTP/2 that is not
    # RSTing the stream correctly when its internal limits are exceeded.
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_01_11_large_resp_headers(self, env: Env, httpd, proto):
        if proto == 'h3' and not env.have_h3():
            pytest.skip("h3 not supported")
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}' \
            f'/curltest/tweak?x-hd={48 * 1024}'
        r = curl.http_get(url=url, alpn_proto=proto, extra_args=[])
        r.check_exit_code(0)
        assert len(r.responses) == 1, f'{r.responses}'
        assert r.responses[0]['status'] == 200, f'{r.responses}'

    # http: response headers larger than what curl buffers for
    @pytest.mark.skipif(condition=not Env.httpd_is_at_least('2.4.64'),
                        reason='httpd must be at least 2.4.64')
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2'])
    def test_01_12_xlarge_resp_headers(self, env: Env, httpd, configures_httpd, proto):
        httpd.set_extra_config('base', [
            f'H2MaxHeaderBlockLen {130 * 1024}',
        ])
        httpd.reload_if_config_changed()
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}' \
            f'/curltest/tweak?x-hd={128 * 1024}'
        r = curl.http_get(url=url, alpn_proto=proto, extra_args=[])
        r.check_exit_code(0)
        assert len(r.responses) == 1, f'{r.responses}'
        assert r.responses[0]['status'] == 200, f'{r.responses}'

    # http: 1 response header larger than what curl buffers for
    @pytest.mark.skipif(condition=not Env.httpd_is_at_least('2.4.64'),
                        reason='httpd must be at least 2.4.64')
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2'])
    def test_01_13_megalarge_resp_headers(self, env: Env, httpd, configures_httpd, proto):
        httpd.set_extra_config('base', [
            'LogLevel http2:trace2',
            f'H2MaxHeaderBlockLen {130 * 1024}',
        ])
        httpd.reload_if_config_changed()
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}' \
            f'/curltest/tweak?x-hd1={128 * 1024}'
        r = curl.http_get(url=url, alpn_proto=proto, extra_args=[])
        if proto == 'h2':
            r.check_exit_code(16)  # CURLE_HTTP2
        else:
            r.check_exit_code(100)  # CURLE_TOO_LARGE

    # http: several response headers, together > 256 KB
    # nghttp2 error -905: Too many CONTINUATION frames following a HEADER frame
    @pytest.mark.skipif(condition=not Env.httpd_is_at_least('2.4.64'),
                        reason='httpd must be at least 2.4.64')
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2'])
    def test_01_14_gigalarge_resp_headers(self, env: Env, httpd, configures_httpd, proto):
        httpd.set_extra_config('base', [
            'LogLevel http2:trace2',
            f'H2MaxHeaderBlockLen {1024 * 1024}',
        ])
        httpd.reload_if_config_changed()
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}' \
            f'/curltest/tweak?x-hd={256 * 1024}'
        r = curl.http_get(url=url, alpn_proto=proto, extra_args=[])
        if proto == 'h2':
            r.check_exit_code(16)  # CURLE_HTTP2
        else:
            r.check_exit_code(0)   # 1.1 can do

    # http: one response header > 256 KB
    @pytest.mark.skipif(condition=not Env.httpd_is_at_least('2.4.64'),
                        reason='httpd must be at least 2.4.64')
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2'])
    def test_01_15_gigalarge_resp_headers(self, env: Env, httpd, configures_httpd, proto):
        httpd.set_extra_config('base', [
            'LogLevel http2:trace2',
            f'H2MaxHeaderBlockLen {1024 * 1024}',
        ])
        httpd.reload_if_config_changed()
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}' \
            f'/curltest/tweak?x-hd1={256 * 1024}'
        r = curl.http_get(url=url, alpn_proto=proto, extra_args=[])
        if proto == 'h2':
            r.check_exit_code(16)  # CURLE_HTTP2
        else:
            r.check_exit_code(100)  # CURLE_TOO_LARGE

    # http: invalid request headers, GET, issue #16998
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_01_16_inv_req_get(self, env: Env, httpd, nghttpx, proto):
        if proto == 'h3' and not env.have_h3():
            pytest.skip("h3 not supported")
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}/curltest/echo'
        r = curl.http_get(url=url, alpn_proto=proto, extra_args=[
            '-H', "a: a\x0ab"
        ])
        # on h1, request is sent, h2/h3 reject
        if proto == 'http/1.1':
            r.check_exit_code(0)
        else:
            r.check_exit_code(43)

    # http: special handling of TE request header
    @pytest.mark.parametrize("te_in, te_out", [
        pytest.param('trailers', 'trailers', id='trailers'),
        pytest.param('chunked', None, id='chunked'),
        pytest.param('gzip, trailers', 'trailers', id='gzip+trailers'),
        pytest.param('gzip ;q=0.2;x="y,x", trailers', 'trailers', id='gzip+q+x+trailers'),
        pytest.param('gzip ;x="trailers", chunks', None, id='gzip+x+chunks'),
    ])
    def test_01_17_TE(self, env: Env, httpd, te_in, te_out):
        proto = 'h2'
        curl = CurlClient(env=env)
        url = f'https://{env.authority_for(env.domain1, proto)}/curltest/echo'
        r = curl.http_download(urls=[url], alpn_proto=proto, with_stats=True,
                               with_headers=True,
                               extra_args=['-H', f'TE: {te_in}'])
        r.check_response(200)
        if te_out is not None:
            assert r.responses[0]['header']['request-te'] == te_out, f'{r.responses[0]}'
        else:
            assert 'request-te' not in r.responses[0]['header'], f'{r.responses[0]}'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  Project                     ___| | | |  _ \| |
#                             / __| | | | |_) | |
#                            | (__| |_| |  _ <| |___
#                             \___|\___/|_| \_\_____|
#
# Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://curl.se/docs/copyright.html.
#
# You may opt to use, copy, modify, merge, publish, distribute and/or sell
# copies of the Software, and permit persons to whom the Software is
# furnished to do so, under the terms of the COPYING file.
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
# KIND, either express or implied.
#
# SPDX-License-Identifier: curl
#
"""A telnet server which negotiates."""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import argparse
import logging
import os
import socket
import sys

from util import ClosingFileHandler

if sys.version_info.major >= 3:
    import socketserver
else:
    import SocketServer as socketserver

log = logging.getLogger(__name__)
HOST = "localhost"
IDENT = "NTEL"


# The strings that indicate the test framework is checking our aliveness
VERIFIED_REQ = "verifiedserver"
VERIFIED_RSP = "WE ROOLZ: {pid}"


def telnetserver(options):
    """Start up a TCP server with a telnet handler and serve DICT requests forever."""
    if options.pidfile:
        pid = os.getpid()
        # see tests/server/util.c function write_pidfile
        if os.name == "nt":
            pid += 4194304
        with open(options.pidfile, "w") as f:
            f.write(str(pid))

    local_bind = (HOST, options.port)
    log.info("Listening on %s", local_bind)

    # Need to set the allow_reuse on the class, not on the instance.
    socketserver.TCPServer.allow_reuse_address = True
    with socketserver.TCPServer(local_bind, NegotiatingTelnetHandler) as server:
        server.serve_forever()
    # leaving `with` calls server.close() automatically
    return ScriptRC.SUCCESS


class NegotiatingTelnetHandler(socketserver.BaseRequestHandler):
    """Handler class for Telnet connections."""

    def handle(self):
        """Negotiates options before reading data."""
        neg = Negotiator(self.request)

        try:
            # Send some initial negotiations.
            neg.send_do("NEW_ENVIRON")
            neg.send_will("NEW_ENVIRON")
            neg.send_dont("NAWS")
            neg.send_wont("NAWS")

            # Get the data passed through the negotiator
            data = neg.recv(4*1024)
            log.debug("Incoming data: %r", data)

            if VERIFIED_REQ.encode('utf-8') in data:
                log.debug("Received verification request from test framework")
                pid = os.getpid()
                # see tests/server/util.c function write_pidfile
                if os.name == "nt":
                    pid += 4194304
                response = VERIFIED_RSP.format(pid=pid)
                response_data = response.encode('utf-8')
            else:
                log.debug("Received normal request - echoing back")
                response_data = data.decode('utf-8').strip().encode('utf-8')

            if response_data:
                log.debug("Sending %r", response_data)
                self.request.sendall(response_data)

            # put some effort into making a clean socket shutdown
            # that does not give the client ECONNRESET
            self.request.settimeout(0.1)
            self.request.recv(4*1024)
            self.request.shutdown(socket.SHUT_RDWR)

        except IOError:
            log.exception("IOError hit during request")


class Negotiator(object):
    NO_NEG = 0
    START_NEG = 1
    WILL = 2
    WONT = 3
    DO = 4
    DONT = 5

    def __init__(self, tcp):
        self.tcp = tcp
        self.state = self.NO_NEG

    def recv(self, bytes):
        """
        Read bytes from TCP, handling negotiation sequences.

        :param bytes: Number of bytes to read
        :return: a buffer of bytes
        """
        buffer = bytearray()

        # If we keep receiving negotiation sequences, we won't fill the buffer.
        # Keep looping while we can, and until we have something to give back
        # to the caller.
        while len(buffer) == 0:
            data = self.tcp.recv(bytes)
            if not data:
                # TCP failed to give us any data. Break out.
                break

            for byte_int in bytearray(data):
                if self.state == self.NO_NEG:
                    self.no_neg(byte_int, buffer)
                elif self.state == self.START_NEG:
                    self.start_neg(byte_int)
                elif self.state in [self.WILL, self.WONT, self.DO, self.DONT]:
                    self.handle_option(byte_int)
                else:
                    # Received an unexpected byte. Stop negotiations
                    log.error("Unexpected byte %s in state %s",
                              byte_int,
                              self.state)
                    self.state = self.NO_NEG

        return buffer

    def no_neg(self, byte_int, buffer):
        # Not negotiating anything thus far. Check to see if we
        # should.
        if byte_int == NegTokens.IAC:
            # Start negotiation
            log.debug("Starting negotiation (IAC)")
            self.state = self.START_NEG
        else:
            # Just append the incoming byte to the buffer
            buffer.append(byte_int)

    def start_neg(self, byte_int):
        # In a negotiation.
        log.debug("In negotiation (%s)",
                  NegTokens.from_val(byte_int))

        if byte_int == NegTokens.WILL:
            # Client is confirming they are willing to do an option
            log.debug("Client is willing")
            self.state = self.WILL
        elif byte_int == NegTokens.WONT:
            # Client is confirming they are unwilling to do an
            # option
            log.debug("Client is unwilling")
            self.state = self.WONT
        elif byte_int == NegTokens.DO:
            # Client is indicating they can do an option
            log.debug("Client can do")
            self.state = self.DO
        elif byte_int == NegTokens.DONT:
            # Client is indicating they can't do an option
            log.debug("Client can't do")
            self.state = self.DONT
        else:
            # Received an unexpected byte. Stop negotiations
            log.error("Unexpected byte %s in state %s",
                      byte_int,
                      self.state)
            self.state = self.NO_NEG

    def handle_option(self, byte_int):
        if byte_int in [NegOptions.BINARY,
                        NegOptions.CHARSET,
                        NegOptions.SUPPRESS_GO_AHEAD,
                        NegOptions.NAWS,
                        NegOptions.NEW_ENVIRON]:
            log.debug("Option: %s", NegOptions.from_val(byte_int))

            # No further negotiation of this option needed. Reset the state.
            self.state = self.NO_NEG

        else:
            # Received an unexpected byte. Stop negotiations
            log.error("Unexpected byte %s in state %s",
                      byte_int,
                      self.state)
            self.state = self.NO_NEG

    def send_message(self, message_ints):
        self.tcp.sendall(bytearray(message_ints))

    def send_iac(self, arr):
        message = [NegTokens.IAC]
        message.extend(arr)
        self.send_message(message)

    def send_do(self, option_str):
        log.debug("Sending DO %s", option_str)
        self.send_iac([NegTokens.DO, NegOptions.to_val(option_str)])

    def send_dont(self, option_str):
        log.debug("Sending DONT %s", option_str)
        self.send_iac([NegTokens.DONT, NegOptions.to_val(option_str)])

    def send_will(self, option_str):
        log.debug("Sending WILL %s", option_str)
        self.send_iac([NegTokens.WILL, NegOptions.to_val(option_str)])

    def send_wont(self, option_str):
        log.debug("Sending WONT %s", option_str)
        self.send_iac([NegTokens.WONT, NegOptions.to_val(option_str)])


class NegBase(object):
    @classmethod
    def to_val(cls, name):
        return getattr(cls, name)

    @classmethod
    def from_val(cls, val):
        for k in cls.__dict__:
            if getattr(cls, k) == val:
                return k

        return "<unknown>"


class NegTokens(NegBase):
    # The start of a negotiation sequence
    IAC = 255
    # Confirm willingness to negotiate
    WILL = 251
    # Confirm unwillingness to negotiate
    WONT = 252
    # Indicate willingness to negotiate
    DO = 253
    # Indicate unwillingness to negotiate
    DONT = 254

    # The start of sub-negotiation options.
    SB = 250
    # The end of sub-negotiation options.
    SE = 240


class NegOptions(NegBase):
    # Binary Transmission
    BINARY = 0
    # Suppress Go Ahead
    SUPPRESS_GO_AHEAD = 3
    # NAWS - width and height of client
    NAWS = 31
    # NEW-ENVIRON - environment variables on client
    NEW_ENVIRON = 39
    # Charset option
    CHARSET = 42


def get_options():
    parser = argparse.ArgumentParser()

    parser.add_argument("--port", action="store", default=9019,
                        type=int, help="port to listen on")
    parser.add_argument("--verbose", action="store", type=int, default=0,
                        help="verbose output")
    parser.add_argument("--pidfile", action="store",
                        help="file name for the PID")
    parser.add_argument("--logfile", action="store",
                        help="file name for the log")
    parser.add_argument("--srcdir", action="store", help="test directory")
    parser.add_argument("--id", action="store", help="server ID")
    parser.add_argument("--ipv4", action="store_true", default=0,
                        help="IPv4 flag")

    return parser.parse_args()


def setup_logging(options):
    """Set up logging from the command line options."""
    root_logger = logging.getLogger()
    add_stdout = False

    formatter = logging.Formatter("%(asctime)s %(levelname)-5.5s "
                                  "[{ident}] %(message)s"
                                  .format(ident=IDENT))

    # Write out to a logfile
    if options.logfile:
        handler = ClosingFileHandler(options.logfile)
        handler.setFormatter(formatter)
        handler.setLevel(logging.DEBUG)
        root_logger.addHandler(handler)
    else:
        # The logfile wasn't specified. Add a stdout logger.
        add_stdout = True

    if options.verbose:
        # Add a stdout logger as well in verbose mode
        root_logger.setLevel(logging.DEBUG)
        add_stdout = True
    else:
        root_logger.setLevel(logging.INFO)

    if add_stdout:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(formatter)
        stdout_handler.setLevel(logging.DEBUG)
        root_logger.addHandler(stdout_handler)


class ScriptRC(object):
    """Enum for script return codes."""

    SUCCESS = 0
    FAILURE = 1
    EXCEPTION = 2


if __name__ == '__main__':
    # Get the options from the user.
    options = get_options()

    # Setup logging using the user options
    setup_logging(options)

    # Run main script.
    try:
        rc = telnetserver(options)
    except Exception:
        log.exception('Error in telnet server')
        rc = ScriptRC.EXCEPTION

    if options.pidfile and os.path.isfile(options.pidfile):
        os.unlink(options.pidfile)

    log.info("Returning %d", rc)
    sys.exit(rc)



#!/usr/bin/env python
"""Compatibility tests between last released and the current version.

Usage: ./update_compatibility_version.py <MAJOR>.<MINOR>.<MICRO> [<RC version>]
Example: ./update_compatibility_version.py 3.7.1
"""

from __future__ import print_function
import re
import sys

if len(sys.argv) < 2 or len(sys.argv) > 3:
  print("""
[ERROR] Please specify a version.

./update_compatibility_version.py.py <MAJOR>.<MINOR>.<MICRO> [<RC version>]

Example:
./update_compatibility_version.py.py 3.7.1 2
""")
  exit(1)

NEW_VERSION = sys.argv[1]
NEW_VERSION_INFO = NEW_VERSION.split('.')
if len(NEW_VERSION_INFO) != 3:
  print("""
[ERROR] Version must be in the format <MAJOR>.<MINOR>.<MICRO>

Example:
./update_compatibility_version.py.py 3.7.3
""")
  exit(1)

if len(sys.argv) > 2:
  RC_VERSION = int(sys.argv[2])
  # Do not update compatibility versions for rc release
  if RC_VERSION != 0:
    exit(0)


def RewriteTextFile(filename, line_rewriter):
  lines = open(filename, 'r').readlines()
  updated_lines = []
  for line in lines:
    updated_lines.append(line_rewriter(line))
  if lines == updated_lines:
    print('%s was not updated. Please double check.' % filename)
  f = open(filename, 'w')
  f.write(''.join(updated_lines))
  f.close()


def ReplaceVersion(line):
  return re.sub(r'LAST_RELEASED=.*$', 'LAST_RELEASED=%s' % NEW_VERSION, line)

RewriteTextFile('tests.sh', ReplaceVersion)



from __future__ import print_function
from __future__ import absolute_import
import argparse
import os
import re
import copy
import uuid
import calendar
import time
import datetime

from util import big_query_utils
from util import result_parser

_PROJECT_ID = 'grpc-testing'
_DATASET = 'protobuf_benchmark_result'
_TABLE = 'opensource_result_v2'
_NOW = "%d%02d%02d" % (datetime.datetime.now().year,
                       datetime.datetime.now().month,
                       datetime.datetime.now().day)

_INITIAL_TIME = calendar.timegm(time.gmtime())

def get_metadata():
  build_number = os.getenv('BUILD_NUMBER')
  build_url = os.getenv('BUILD_URL')
  job_name = os.getenv('JOB_NAME')
  git_commit = os.getenv('GIT_COMMIT')
  # actual commit is the actual head of PR that is getting tested
  git_actual_commit = os.getenv('ghprbActualCommit')

  utc_timestamp = str(calendar.timegm(time.gmtime()))
  metadata = {'created': utc_timestamp}

  if build_number:
    metadata['buildNumber'] = build_number
  if build_url:
    metadata['buildUrl'] = build_url
  if job_name:
    metadata['jobName'] = job_name
  if git_commit:
    metadata['gitCommit'] = git_commit
  if git_actual_commit:
    metadata['gitActualCommit'] = git_actual_commit

  return metadata


def upload_result(result_list, metadata):
  for result in result_list:
    new_result = {}
    new_result["metric"] = "throughput"
    new_result["value"] = result["throughput"]
    new_result["unit"] = "MB/s"
    new_result["test"] = "protobuf_benchmark"
    new_result["product_name"] = "protobuf"
    labels_string = ""
    for key in result:
      labels_string += ",|%s:%s|" % (key, result[key])
    new_result["labels"] = labels_string[1:]
    new_result["timestamp"] = _INITIAL_TIME
    print(labels_string)

    bq = big_query_utils.create_big_query()
    row = big_query_utils.make_row(str(uuid.uuid4()), new_result)
    if not big_query_utils.insert_rows(bq, _PROJECT_ID, _DATASET,
                                        _TABLE + "$" + _NOW,
                                        [row]):
      print('Error when uploading result', new_result)


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("-cpp", "--cpp_input_file",
                      help="The CPP benchmark result file's name",
                      default="")
  parser.add_argument("-java", "--java_input_file",
                      help="The Java benchmark result file's name",
                      default="")
  parser.add_argument("-python", "--python_input_file",
                      help="The Python benchmark result file's name",
                      default="")
  parser.add_argument("-go", "--go_input_file",
                      help="The golang benchmark result file's name",
                      default="")
  parser.add_argument("-node", "--node_input_file",
                      help="The node.js benchmark result file's name",
                      default="")
  parser.add_argument("-php", "--php_input_file",
                      help="The pure php benchmark result file's name",
                      default="")
  parser.add_argument("-php_c", "--php_c_input_file",
                      help="The php with c ext benchmark result file's name",
                      default="")
  args = parser.parse_args()

  metadata = get_metadata()
  print("uploading results...")
  upload_result(result_parser.get_result_from_file(
      cpp_file=args.cpp_input_file,
      java_file=args.java_input_file,
      python_file=args.python_input_file,
      go_file=args.go_input_file,
      node_file=args.node_input_file,
      php_file=args.php_input_file,
      php_c_file=args.php_c_input_file,
  ), metadata)



# Copyright 2009 Baptiste Lepilleur and The JsonCpp Authors
# Distributed under MIT license, or public domain if desired and
# recognized in your jurisdiction.
# See file LICENSE for detail or copy at http://jsoncpp.sourceforge.net/LICENSE

from __future__ import print_function
from __future__ import unicode_literals
from io import open
from glob import glob
import sys
import os
import os.path
import subprocess
import optparse

VALGRIND_CMD = 'valgrind --tool=memcheck --leak-check=yes --undef-value-errors=yes'

class TestProxy(object):
    def __init__(self, test_exe_path, use_valgrind=False):
        self.test_exe_path = os.path.normpath(os.path.abspath(test_exe_path))
        self.use_valgrind = use_valgrind

    def run(self, options):
        if self.use_valgrind:
            cmd = VALGRIND_CMD.split()
        else:
            cmd = []
        cmd.extend([self.test_exe_path, '--test-auto'] + options)
        try:
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except:
            print(cmd)
            raise
        stdout = process.communicate()[0]
        if process.returncode:
            return False, stdout
        return True, stdout

def runAllTests(exe_path, use_valgrind=False):
    test_proxy = TestProxy(exe_path, use_valgrind=use_valgrind)
    status, test_names = test_proxy.run(['--list-tests'])
    if not status:
        print("Failed to obtain unit tests list:\n" + test_names, file=sys.stderr)
        return 1
    test_names = [name.strip() for name in test_names.decode('utf-8').strip().split('\n')]
    failures = []
    for name in test_names:
        print('TESTING %s:' % name, end=' ')
        succeed, result = test_proxy.run(['--test', name])
        if succeed:
            print('OK')
        else:
            failures.append((name, result))
            print('FAILED')
    failed_count = len(failures)
    pass_count = len(test_names) - failed_count
    if failed_count:
        print()
        for name, result in failures:
            print(result)
        print('%d/%d tests passed (%d failure(s))' % (            pass_count, len(test_names), failed_count))
        return 1
    else:
        print('All %d tests passed' % len(test_names))
        return 0

def main():
    from optparse import OptionParser
    parser = OptionParser(usage="%prog [options] <path to test_lib_json.exe>")
    parser.add_option("--valgrind",
                  action="store_true", dest="valgrind", default=False,
                  help="run all the tests using valgrind to detect memory leaks")
    parser.enable_interspersed_args()
    options, args = parser.parse_args()

    if len(args) != 1:
        parser.error('Must provides at least path to test_lib_json executable.')
        sys.exit(1)

    exit_code = runAllTests(args[0], use_valgrind=options.valgrind)
    sys.exit(exit_code)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""Generates a friendly list of changes per language since the last release."""

import sys
import os

class Language(object):
  def __init__(self, name, pathspec):
    self.name = name
    self.pathspec = pathspec

languages = [
  Language("C++", [
      "':(glob)src/google/protobuf/*'",
      "src/google/protobuf/compiler/cpp",
      "src/google/protobuf/io",
      "src/google/protobuf/util",
      "src/google/protobuf/stubs",
  ]),
  Language("Java", [
      "java",
      "src/google/protobuf/compiler/java",
  ]),
  Language("Python", [
      "python",
      "src/google/protobuf/compiler/python",
  ]),
  Language("PHP", [
      "php",
      "src/google/protobuf/compiler/php",
  ]),
  Language("Ruby", [
      "ruby",
      "src/google/protobuf/compiler/ruby",
  ]),
  Language("Csharp", [
      "csharp",
      "src/google/protobuf/compiler/csharp",
  ]),
  Language("Objective C", [
      "objectivec",
      "src/google/protobuf/compiler/objectivec",
  ]),
]

if len(sys.argv) < 2:
  print("Usage: generate_changelog.py <previous release>")
  sys.exit(1)

previous = sys.argv[1]

for language in languages:
  print(language.name)
  sys.stdout.flush()
  os.system(("git log --pretty=oneline --abbrev-commit %s...HEAD %s | " +
             "sed -e 's/^/ - /'") % (previous, " ".join(language.pathspec)))
  print("")

print("To view a commit on GitHub: " +
      "https://github.com/protocolbuffers/protobuf/commit/<commit id>")



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
#
# Copyright (C) 2006-2026 by The Odamex Team.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# DESCRIPTION:
#  Updates the version number and copyright date on a given set of source
#  files.  This is a version that is less safe and assumes a prior version
#  to make up the difference in safety.
#
# -----------------------------------------------------------------------------

from typing import List, Tuple

import re, sys, os
from configparser import ConfigParser
from pathlib import Path

SCRIPT_DIR = Path(__file__).resolve().parent
ROOT_DIR = Path(SCRIPT_DIR.parent.parent)


def re_replace(pattern: str, regex: str, repl: str):
    """
    Replace a pattern in a file.
    """

    for inode in ROOT_DIR.glob(pattern):
        with open(inode, "rb") as f:
            contents = f.read().decode("utf8", errors="surrogateescape")
            (contents, count) = re.subn(regex, repl, contents)
        if count > 0:
            with open(inode, "wb") as f:
                f.write(contents.encode("utf8", errors="surrogateescape"))
            print(f'Wrote "{inode}".')
        else:
            print(f'No matches found in "{inode}", skipping...')


def savestr(major: int, minor: int, patch: int) -> str:
    """
    Turn a version into a save-string.

    This string is exactly 16 characters and used as the header of the save
    file.  This system automatically flexes from single-digit numbers into
    taking up all 16 characters when we escape single digits to avoid ambiguous
    version headers.
    """

    if major <= 9 and minor <= 9 and patch <= 9:
        # We can do single-digit maj/min/patch.
        buffer = f"ODAMEXSAVE{major}{minor}{patch}   "
    else:
        # Patch can only be 0-9, Minor can only be 1-25.
        buffer = f"ODAMEXSAVE{major:03}{minor:02}{patch:01}"

    assert len(buffer) == 16
    return buffer


def configstr(major: int, minor: int, patch: int) -> str:
    """
    Turn a version into a config-string.

    This is hoisted into the config as a string, but is intended to be read
    as a number.  Numbers for newer versions must always compare greater
    than older versions when the string is turned into an int.
    """

    if major == 0 and minor <= 9 and patch <= 9:
        # We can do single-digit min/patch.
        buffer = f"{minor}{patch}"
    else:
        # Patch can only be 0-9, Minor can only be 1-25.
        buffer = f"{major:03}{minor:02}{patch:01}"

    return buffer


def ini_version(input: str) -> Tuple[int, int, int]:
    ver = [int(x) for x in input.split(".")]
    assert len(ver) == 3
    return tuple(ver)


def ini_files(input: str) -> List[str]:
    files = input.split("\n")
    return [x for x in files if len(x) > 0]


def main():
    with open(SCRIPT_DIR / "upversion.ini", "r") as fh:
        config = ConfigParser()
        config.read_file(fh)

    if not os.getenv('OLD_VERSION'):
        OLD_VERSION = "1.0.0"
    else:
        OLD_VERSION = os.getenv('OLD_VERSION')

    if not os.getenv('NEW_VERSION'):
        NEW_VERSION = "2.0.0"
    else:
        NEW_VERSION = os.getenv('NEW_VERSION')

    if not os.getenv('OLD_YEAR'):
        OLD_YEAR = r"2006-2026"
    else:
        OLD_YEAR = os.getenv('OLD_YEAR')

    if not os.getenv('NEW_YEAR'):
        NEW_YEAR = r"2006-2026"
    else:
        NEW_YEAR = os.getenv('NEW_YEAR')

    DOTTED_FILES = ini_files(config.get("files", "dotted"))
    COMMA_FILES = ini_files(config.get("files", "comma"))
    SAVESTR_FILES = ini_files(config.get("files", "savestr"))
    CONFIGSTR_FILES = ini_files(config.get("files", "configstr"))
    PLAINYEAR_FILES = ini_files(config.get("files", "plainyear"))
    FANCYYEAR_FILES = ini_files(config.get("files", "fancyyear"))

    oldmaj = int(OLD_VERSION.split(".")[0])
    oldmin = int(OLD_VERSION.split(".")[1])
    oldpatch = int(OLD_VERSION.split(".")[2])

    newmaj = int(NEW_VERSION.split(".")[0])
    newmin = int(NEW_VERSION.split(".")[1])
    newpatch = int(NEW_VERSION.split(".")[2])

    # Dotted version.
    DOTTED_RE = f"{oldmaj}\\.{oldmin}\\.{oldpatch}"
    DOTTED_REPL = f"{newmaj}.{newmin}.{newpatch}"
    [re_replace(glob, DOTTED_RE, DOTTED_REPL) for glob in DOTTED_FILES]

    # Comma-separated version.
    COMMA_RE = f"{oldmaj}, {oldmin}, {oldpatch}"
    COMMA_REPL = f"{newmaj}, {newmin}, {newpatch}"
    [re_replace(glob, COMMA_RE, COMMA_REPL) for glob in COMMA_FILES]

    # Fixed-length save strings.
    [
        re_replace(glob, savestr(oldmaj, oldmin, oldpatch), savestr(newmaj, newmin, newpatch))
        for glob in SAVESTR_FILES
    ]

    # Config strings.
    [
        re_replace(glob, configstr(oldmaj, oldmin, oldpatch), configstr(newmaj, newmin, newpatch))
        for glob in CONFIGSTR_FILES
    ]

    # CMake PROJECT_RC_VERSION
    CMAKERC_RE = f'"{oldmaj},{oldmin},{oldpatch},0"'
    CMAKERC_REPL = f'"{newmaj},{newmin},{newpatch},0"'
    re_replace("CMakeLists.txt", CMAKERC_RE, CMAKERC_REPL)

    # CMake PROJECT_COPYRIGHT
    re_replace("CMakeLists.txt", f'"{OLD_YEAR}"', f'"{NEW_YEAR}"')

    # Plain old copyright replacement - avoids comments.
    PLAINYEAR_RE = f"([^ ])Copyright \\(C\\) {OLD_YEAR} The Odamex Team([^ ])"
    PLAINYEAR_REPL = f"\\1Copyright (C) {NEW_YEAR} The Odamex Team\\2"
    [re_replace(glob, PLAINYEAR_RE, PLAINYEAR_REPL) for glob in PLAINYEAR_FILES]

    # Fancy copyright replacement - avoids comments.
    FANCYYEAR_RE = f"Copyright © {OLD_YEAR} The Odamex Team"
    FANCYYEAR_REPL = f"Copyright © {NEW_YEAR} The Odamex Team"
    [re_replace(glob, FANCYYEAR_RE, FANCYYEAR_REPL) for glob in FANCYYEAR_FILES]


if __name__ == "__main__":
    main()



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
# https://developers.google.com/protocol-buffers/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Setuptools/distutils extension for generating Python protobuf code."""

__author__ = 'dlj@google.com (David L. Jones)'

from os import path
from setuptools import setup, find_packages

# Use README.md as the source for long_description.
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    _readme = f.read()

setup(
    name='protobuf_distutils',
    version='1.0',
    packages=find_packages(),
    maintainer='protobuf@googlegroups.com',
    maintainer_email='protobuf@googlegroups.com',
    license='BSD-3-Clause',
    classifiers=[
        "Framework :: Setuptools Plugin",
        "Operating System :: OS Independent",
        # These Python versions should match the protobuf package:
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Code Generators",
    ],
    description=('This is a distutils extension to generate Python code for '
                 '.proto files using an installed protoc binary.'),
    long_description=_readme,
    long_description_content_type='text/markdown',
    url='https://github.com/protocolbuffers/protobuf/',
    entry_points={
        'distutils.commands': [
            ('generate_py_protobufs = '
             'protobuf_distutils.generate_py_protobufs:generate_py_protobufs'),
        ],
    },
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#***************************************************************************
#                                  _   _ ____  _
#  Project                     ___| | | |  _ \| |
#                             / __| | | | |_) | |
#                            | (__| |_| |  _ <| |___
#                             \___|\___/|_| \_\_____|
#
# Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://curl.se/docs/copyright.html.
#
# You may opt to use, copy, modify, merge, publish, distribute and/or sell
# copies of the Software, and permit persons to whom the Software is
# furnished to do so, under the terms of the COPYING file.
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
# KIND, either express or implied.
#
# SPDX-License-Identifier: curl
#
###########################################################################
#
import difflib
import filecmp
import logging
import os
import re
import pytest

from testenv import Env, CurlClient, Caddy, LocalClient


log = logging.getLogger(__name__)


@pytest.mark.skipif(condition=not Env.has_caddy(), reason="missing caddy")
@pytest.mark.skipif(condition=not Env.have_ssl_curl(), reason="curl without SSL")
class TestCaddy:

    @pytest.fixture(autouse=True, scope='class')
    def caddy(self, env):
        caddy = Caddy(env=env)
        assert caddy.initial_start()
        yield caddy
        caddy.stop()

    def _make_docs_file(self, docs_dir: str, fname: str, fsize: int):
        fpath = os.path.join(docs_dir, fname)
        data1k = 1024*'x'
        flen = 0
        with open(fpath, 'w') as fd:
            while flen < fsize:
                fd.write(data1k)
                flen += len(data1k)
        return flen

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env, caddy):
        self._make_docs_file(docs_dir=caddy.docs_dir, fname='data10k.data', fsize=10*1024)
        self._make_docs_file(docs_dir=caddy.docs_dir, fname='data1.data', fsize=1024*1024)
        self._make_docs_file(docs_dir=caddy.docs_dir, fname='data5.data', fsize=5*1024*1024)
        self._make_docs_file(docs_dir=caddy.docs_dir, fname='data10.data', fsize=10*1024*1024)
        self._make_docs_file(docs_dir=caddy.docs_dir, fname='data100.data', fsize=100*1024*1024)
        env.make_data_file(indir=env.gen_dir, fname="data-10m", fsize=10*1024*1024)

    # download 1 file
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_08_01_download_1(self, env: Env, caddy: Caddy, proto):
        if proto == 'h3' and not env.have_h3_curl():
            pytest.skip("h3 not supported in curl")
        curl = CurlClient(env=env)
        url = f'https://{env.domain1}:{caddy.port}/data.json'
        r = curl.http_download(urls=[url], alpn_proto=proto)
        r.check_response(count=1, http_status=200)

    # download 1MB files sequentially
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_08_02_download_1mb_sequential(self, env: Env, caddy: Caddy, proto):
        if proto == 'h3' and not env.have_h3_curl():
            pytest.skip("h3 not supported in curl")
        count = 50
        curl = CurlClient(env=env)
        urln = f'https://{env.domain1}:{caddy.port}/data1.data?[0-{count-1}]'
        r = curl.http_download(urls=[urln], alpn_proto=proto)
        r.check_response(count=count, http_status=200, connect_count=1)

    # download 1MB files parallel
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_08_03_download_1mb_parallel(self, env: Env, caddy: Caddy, proto):
        if proto == 'h3' and not env.have_h3_curl():
            pytest.skip("h3 not supported in curl")
        count = 20
        curl = CurlClient(env=env)
        urln = f'https://{env.domain1}:{caddy.port}/data1.data?[0-{count-1}]'
        r = curl.http_download(urls=[urln], alpn_proto=proto, extra_args=[
            '--parallel'
        ])
        r.check_response(count=count, http_status=200)
        if proto == 'http/1.1':
            # http/1.1 parallel transfers will open multiple connections
            assert r.total_connects > 1, r.dump_logs()
        else:
            assert r.total_connects == 1, r.dump_logs()

    # download 5MB files sequentially
    @pytest.mark.skipif(condition=Env().slow_network, reason="not suitable for slow network tests")
    @pytest.mark.skipif(condition=Env().ci_run, reason="not suitable for CI runs")
    @pytest.mark.parametrize("proto", ['h2', 'h3'])
    def test_08_04a_download_10mb_sequential(self, env: Env, caddy: Caddy, proto):
        if proto == 'h3' and not env.have_h3_curl():
            pytest.skip("h3 not supported in curl")
        count = 40
        curl = CurlClient(env=env)
        urln = f'https://{env.domain1}:{caddy.port}/data5.data?[0-{count-1}]'
        r = curl.http_download(urls=[urln], alpn_proto=proto)
        r.check_response(count=count, http_status=200, connect_count=1)

    # download 10MB files sequentially
    @pytest.mark.skipif(condition=Env().slow_network, reason="not suitable for slow network tests")
    @pytest.mark.skipif(condition=Env().ci_run, reason="not suitable for CI runs")
    @pytest.mark.parametrize("proto", ['h2', 'h3'])
    def test_08_04b_download_10mb_sequential(self, env: Env, caddy: Caddy, proto):
        if proto == 'h3' and not env.have_h3_curl():
            pytest.skip("h3 not supported in curl")
        count = 20
        curl = CurlClient(env=env)
        urln = f'https://{env.domain1}:{caddy.port}/data10.data?[0-{count-1}]'
        r = curl.http_download(urls=[urln], alpn_proto=proto)
        r.check_response(count=count, http_status=200, connect_count=1)

    # download 10MB files parallel
    @pytest.mark.skipif(condition=Env().slow_network, reason="not suitable for slow network tests")
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    @pytest.mark.skipif(condition=Env().ci_run, reason="not suitable for CI runs")
    def test_08_05_download_1mb_parallel(self, env: Env, caddy: Caddy, proto):
        if proto == 'h3' and not env.have_h3_curl():
            pytest.skip("h3 not supported in curl")
        if proto == 'http/1.1' and env.curl_uses_lib('mbedtls'):
            pytest.skip("mbedtls 3.6.0 fails on 50 connections with: "
                        "ssl_handshake returned: (-0x7F00) SSL - Memory allocation failed")
        count = 50
        curl = CurlClient(env=env)
        urln = f'https://{env.domain1}:{caddy.port}/data10.data?[0-{count-1}]'
        r = curl.http_download(urls=[urln], alpn_proto=proto, extra_args=[
            '--parallel'
        ])
        r.check_response(count=count, http_status=200)
        if proto == 'http/1.1':
            # http/1.1 parallel transfers will open multiple connections
            assert r.total_connects > 1, r.dump_logs()
        else:
            assert r.total_connects == 1, r.dump_logs()

    # post data parallel, check that they were echoed
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_08_06_post_parallel(self, env: Env, httpd, caddy, proto):
        if proto == 'h3' and not env.have_h3():
            pytest.skip("h3 not supported")
        # limit since we use a separate connection in h1
        count = 20
        data = '0123456789'
        curl = CurlClient(env=env)
        url = f'https://{env.domain2}:{caddy.port}/curltest/echo?id=[0-{count-1}]'
        r = curl.http_upload(urls=[url], data=data, alpn_proto=proto,
                             extra_args=['--parallel'])
        r.check_stats(count=count, http_status=200, exitcode=0)
        for i in range(count):
            respdata = open(curl.response_file(i)).readlines()
            assert respdata == [data]

    # put large file, check that they length were echoed
    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_08_07_put_large(self, env: Env, httpd, caddy, proto):
        if proto == 'h3' and not env.have_h3():
            pytest.skip("h3 not supported")
        # limit since we use a separate connection in h1<
        count = 1
        fdata = os.path.join(env.gen_dir, 'data-10m')
        curl = CurlClient(env=env)
        url = f'https://{env.domain2}:{caddy.port}/curltest/put?id=[0-{count-1}]'
        r = curl.http_put(urls=[url], fdata=fdata, alpn_proto=proto)
        exp_data = [f'{os.path.getsize(fdata)}']
        r.check_response(count=count, http_status=200)
        for i in range(count):
            respdata = open(curl.response_file(i)).readlines()
            assert respdata == exp_data

    @pytest.mark.parametrize("proto", ['http/1.1', 'h2', 'h3'])
    def test_08_08_earlydata(self, env: Env, httpd, caddy, proto):
        if not env.curl_can_early_data():
            pytest.skip('TLS earlydata not implemented')
        if proto == 'h3' and \
           (not env.have_h3() or not env.curl_can_h3_early_data()):
            pytest.skip("h3 not supported")
        count = 2
        docname = 'data10k.data'
        url = f'https://{env.domain1}:{caddy.port}/{docname}'
        client = LocalClient(name='cli_hx_download', env=env)
        if not client.exists():
            pytest.skip(f'example client not built: {client.name}')
        r = client.run(args=[
             '-n', f'{count}',
             '-e',  # use TLS earlydata
             '-f',  # forbid reuse of connections
             '-r', f'{env.domain1}:{caddy.port}:127.0.0.1',
             '-V', proto, url
        ])
        r.check_exit_code(0)
        srcfile = os.path.join(caddy.docs_dir, docname)
        self.check_downloads(client, srcfile, count)
        earlydata = {}
        for line in r.trace_lines:
            m = re.match(r'^\[t-(\d+)] EarlyData: (-?\d+)', line)
            if m:
                earlydata[int(m.group(1))] = int(m.group(2))
        assert earlydata[0] == 0, f'{earlydata}'
        if proto == 'h3':
            assert earlydata[1] == 113, f'{earlydata}'
        else:
            # Caddy does not support early data on TCP
            assert earlydata[1] == 0, f'{earlydata}'

    def check_downloads(self, client, srcfile: str, count: int,
                        complete: bool = True):
        for i in range(count):
            dfile = client.download_file(i)
            assert os.path.exists(dfile)
            if complete and not filecmp.cmp(srcfile, dfile, shallow=False):
                diff = "".join(difflib.unified_diff(a=open(srcfile).readlines(),
                                                    b=open(dfile).readlines(),
                                                    fromfile=srcfile,
                                                    tofile=dfile,
                                                    n=1))
                assert False, f'download {dfile} differs:\n{diff}'



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
# https://developers.google.com/protocol-buffers/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Determine which implementation of the protobuf API is used in this process.
"""

import importlib
import os
import sys
import warnings


def _ApiVersionToImplementationType(api_version):
  if api_version == 2:
    return 'cpp'
  if api_version == 1:
    raise ValueError('api_version=1 is no longer supported.')
  if api_version == 0:
    return 'python'
  return None


_implementation_type = None
try:
  # pylint: disable=g-import-not-at-top
  from google.protobuf.internal import _api_implementation
  # The compile-time constants in the _api_implementation module can be used to
  # switch to a certain implementation of the Python API at build time.
  _implementation_type = _ApiVersionToImplementationType(
      _api_implementation.api_version)
except ImportError:
  pass  # Unspecified by compiler flags.


def _CanImport(mod_name):
  try:
    mod = importlib.import_module(mod_name)
    # Work around a known issue in the classic bootstrap .par import hook.
    if not mod:
      raise ImportError(mod_name + ' import succeeded but was None')
    return True
  except ImportError:
    return False


if _implementation_type is None:
  if _CanImport('google._upb._message'):
    _implementation_type = 'upb'
  elif _CanImport('google.protobuf.pyext._message'):
    _implementation_type = 'cpp'
  else:
    _implementation_type = 'python'


# This environment variable can be used to switch to a certain implementation
# of the Python API, overriding the compile-time constants in the
# _api_implementation module. Right now only 'python', 'cpp' and 'upb' are
# valid values. Any other value will raise error.
_implementation_type = os.getenv('PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION',
                                 _implementation_type)

if _implementation_type not in ('python', 'cpp', 'upb'):
  raise ValueError('PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION {0} is not '
                   'supported. Please set to \'python\', \'cpp\' or '
                   '\'upb\'.'.format(_implementation_type))

if 'PyPy' in sys.version and _implementation_type == 'cpp':
  warnings.warn('PyPy does not work yet with cpp protocol buffers. '
                'Falling back to the python implementation.')
  _implementation_type = 'python'

_c_module = None

if _implementation_type == 'cpp':
  try:
    # pylint: disable=g-import-not-at-top
    from google.protobuf.pyext import _message
    _c_module = _message
    del _message
  except ImportError:
    # TODO(jieluo): fail back to python
    warnings.warn(
        'Selected implementation cpp is not available.')
    pass

if _implementation_type == 'upb':
  try:
    # pylint: disable=g-import-not-at-top
    from google._upb import _message
    _c_module = _message
    del _message
  except ImportError:
    warnings.warn('Selected implementation upb is not available. '
                  'Falling back to the python implementation.')
    _implementation_type = 'python'
    pass

# Detect if serialization should be deterministic by default
try:
  # The presence of this module in a build allows the proto implementation to
  # be upgraded merely via build deps.
  #
  # NOTE: Merely importing this automatically enables deterministic proto
  # serialization for C++ code, but we still need to export it as a boolean so
  # that we can do the same for `_implementation_type == 'python'`.
  #
  # NOTE2: It is possible for C++ code to enable deterministic serialization by
  # default _without_ affecting Python code, if the C++ implementation is not in
  # use by this module.  That is intended behavior, so we don't actually expose
  # this boolean outside of this module.
  #
  # pylint: disable=g-import-not-at-top,unused-import
  from google.protobuf import enable_deterministic_proto_serialization
  _python_deterministic_proto_serialization = True
except ImportError:
  _python_deterministic_proto_serialization = False


# Usage of this function is discouraged. Clients shouldn't care which
# implementation of the API is in use. Note that there is no guarantee
# that differences between APIs will be maintained.
# Please don't use this function if possible.
def Type():
  return _implementation_type


def _SetType(implementation_type):
  """Never use! Only for protobuf benchmark."""
  global _implementation_type
  _implementation_type = implementation_type


# See comment on 'Type' above.
# TODO(jieluo): Remove the API, it returns a constant. b/228102101
def Version():
  return 2


# For internal use only
def IsPythonDefaultSerializationDeterministic():
  return _python_deterministic_proto_serialization



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#***************************************************************************
#                                  _   _ ____  _
#  Project                     ___| | | |  _ \| |
#                             / __| | | | |_) | |
#                            | (__| |_| |  _ <| |___
#                             \___|\___/|_| \_\_____|
#
# Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://curl.se/docs/copyright.html.
#
# You may opt to use, copy, modify, merge, publish, distribute and/or sell
# copies of the Software, and permit persons to whom the Software is
# furnished to do so, under the terms of the COPYING file.
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
# KIND, either express or implied.
#
# SPDX-License-Identifier: curl
#
###########################################################################
#
import logging
import os
import shutil
import subprocess
from datetime import datetime
from typing import Optional, Dict

from . import ExecResult
from .env import Env


log = logging.getLogger(__name__)


class LocalClient:

    def __init__(self, name: str, env: Env, run_dir: Optional[str] = None,
                 timeout: Optional[float] = None,
                 run_env: Optional[Dict[str,str]] = None):
        self.name = name
        self.path = os.path.join(env.build_dir, 'tests/libtest/libtests')
        self.env = env
        self._run_env = run_env
        self._timeout = timeout if timeout else env.test_timeout
        self._curl = os.environ['CURL'] if 'CURL' in os.environ else env.curl
        self._run_dir = run_dir if run_dir else os.path.join(env.gen_dir, name)
        self._stdoutfile = f'{self._run_dir}/stdout'
        self._stderrfile = f'{self._run_dir}/stderr'
        self._rmrf(self._run_dir)
        self._mkpath(self._run_dir)

    @property
    def run_dir(self) -> str:
        return self._run_dir

    @property
    def stderr_file(self) -> str:
        return self._stderrfile

    def exists(self) -> bool:
        return os.path.exists(self.path)

    def download_file(self, i: int) -> str:
        return os.path.join(self._run_dir, f'download_{i}.data')

    def _rmf(self, path):
        if os.path.exists(path):
            return os.remove(path)

    def _rmrf(self, path):
        if os.path.exists(path):
            return shutil.rmtree(path)

    def _mkpath(self, path):
        if not os.path.exists(path):
            return os.makedirs(path)

    def run(self, args):
        self._rmf(self._stdoutfile)
        self._rmf(self._stderrfile)
        start = datetime.now()
        exception = None
        myargs = [self.path, self.name]
        myargs.extend(args)
        run_env = None
        if self._run_env:
            run_env = self._run_env.copy()
            for key in ['CURL_DEBUG']:
                if key in os.environ and key not in run_env:
                    run_env[key] = os.environ[key]
        try:
            with open(self._stdoutfile, 'w') as cout, open(self._stderrfile, 'w') as cerr:
                p = subprocess.run(myargs, stderr=cerr, stdout=cout,
                                   cwd=self._run_dir, shell=False,
                                   input=None, env=run_env,
                                   timeout=self._timeout)
                exitcode = p.returncode
        except subprocess.TimeoutExpired:
            log.warning(f'Timeout after {self._timeout}s: {args}')
            exitcode = -1
            exception = 'TimeoutExpired'
        coutput = open(self._stdoutfile).readlines()
        cerrput = open(self._stderrfile).readlines()
        return ExecResult(args=myargs, exit_code=exitcode, exception=exception,
                          stdout=coutput, stderr=cerrput,
                          duration=datetime.now() - start)

    def dump_logs(self):
        lines = []
        lines.append('>>--stdout ----------------------------------------------\n')
        lines.extend(open(self._stdoutfile).readlines())
        lines.append('>>--stderr ----------------------------------------------\n')
        lines.extend(open(self._stderrfile).readlines())
        lines.append('<<-------------------------------------------------------\n')
        return ''.join(lines)



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
# https://developers.google.com/protocol-buffers/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Implements the generate_py_protobufs command."""

__author__ = 'dlj@google.com (David L. Jones)'

import glob
import sys
import os
import distutils.spawn as spawn
from distutils.cmd import Command
from distutils.errors import DistutilsOptionError, DistutilsExecError

class generate_py_protobufs(Command):
    """Generates Python sources for .proto files."""

    description = 'Generate Python sources for .proto files'
    user_options = [
        ('extra-proto-paths=', None,
         'Additional paths to resolve imports in .proto files.'),

        ('protoc=', None,
         'Path to a specific `protoc` command to use.'),
    ]
    boolean_options = ['recurse']

    def initialize_options(self):
        """Sets the defaults for the command options."""
        self.source_dir = None
        self.proto_root_path = None
        self.extra_proto_paths = []
        self.output_dir = '.'
        self.proto_files = None
        self.recurse = True
        self.protoc = None

    def finalize_options(self):
        """Sets the final values for the command options.

        Defaults were set in `initialize_options`, but could have been changed
        by command-line options or by other commands.
        """
        self.ensure_dirname('source_dir')
        self.ensure_string_list('extra_proto_paths')

        if self.output_dir is None:
            self.output_dir = '.'
        self.ensure_dirname('output_dir')

        # SUBTLE: if 'source_dir' is a subdirectory of any entry in
        # 'extra_proto_paths', then in general, the shortest --proto_path prefix
        # (and the longest relative .proto filenames) must be used for
        # correctness. For example, consider:
        #
        #     source_dir = 'a/b/c'
        #     extra_proto_paths = ['a/b', 'x/y']
        #
        # In this case, we must ensure that a/b/c/d/foo.proto resolves
        # canonically as c/d/foo.proto, not just d/foo.proto. Otherwise, this
        # import:
        #
        #     import "c/d/foo.proto";
        #
        # would result in different FileDescriptor.name keys from "d/foo.proto".
        # That will cause all the definitions in the file to be flagged as
        # duplicates, with an error similar to:
        #
        #     c/d/foo.proto: "packagename.MessageName" is already defined in file "d/foo.proto"
        #
        # For paths in self.proto_files, we transform them to be relative to
        # self.proto_root_path, which may be different from self.source_dir.
        #
        # Although the order of --proto_paths is significant, shadowed filenames
        # are errors: if 'a/b/c.proto' resolves to different files under two
        # different --proto_path arguments, then the path is rejected as an
        # error. (Implementation note: this is enforced in protoc's
        # DiskSourceTree class.)

        if self.proto_root_path is None:
            self.proto_root_path = os.path.normpath(self.source_dir)
            for root_candidate in self.extra_proto_paths:
                root_candidate = os.path.normpath(root_candidate)
                if self.proto_root_path.startswith(root_candidate):
                    self.proto_root_path = root_candidate
            if self.proto_root_path != self.source_dir:
                self.announce('using computed proto_root_path: ' + self.proto_root_path, level=2)

        if not self.source_dir.startswith(self.proto_root_path):
            raise DistutilsOptionError('source_dir ' + self.source_dir +
                                       ' is not under proto_root_path ' + self.proto_root_path)

        if self.proto_files is None:
            files = glob.glob(os.path.join(self.source_dir, '*.proto'))
            if self.recurse:
                files.extend(glob.glob(os.path.join(self.source_dir, '**', '*.proto'), recursive=True))
            self.proto_files = [f.partition(self.proto_root_path + os.path.sep)[-1] for f in files]
            if not self.proto_files:
                raise DistutilsOptionError('no .proto files were found under ' + self.source_dir)

        self.ensure_string_list('proto_files')

        if self.protoc is None:
            self.protoc = os.getenv('PROTOC')
        if self.protoc is None:
            self.protoc = spawn.find_executable('protoc')

    def run(self):
        # All proto file paths were adjusted in finalize_options to be relative
        # to self.proto_root_path.
        proto_paths = ['--proto_path=' + self.proto_root_path]
        proto_paths.extend(['--proto_path=' + x for x in self.extra_proto_paths])

        # Run protoc. It was already resolved, so don't try to resolve
        # through PATH.
        spawn.spawn(
            [self.protoc,
             '--python_out=' + self.output_dir,
            ] + proto_paths + self.proto_files,
            search_path=0)



#!/usr/bin/env python3
# vim: sw=4 ts=4 expandtab
# (c) 2021 Thomas BERNARD
# Python3 module to submit service to running MiniSSDPd
# MiniSSDPd: See http://miniupnp.free.fr/minissdpd.html
import socket, os


def codelength(s):
    """ returns the given string/bytes as bytes, prepended with the 7-bit-encoded length """
    # We want bytes
    if not isinstance(s, bytes):
        # Not bytes. Let's try to convert to bytes, but only plain ASCII
        try:
            s = str.encode(s, "ascii")
        except:
            s = b''
    l = len(s)
    if l == 0:
        return b'\x00'
    encodedlen = (l & 0x7F).to_bytes(1, 'little')
    while l > 0x7F:
        l = l >> 7
        c = (l & 0x7F) | 0x80
        encodedlen = c.to_bytes(1, 'little') + encodedlen
    return encodedlen + s


def submit_to_minissdpd(st, usn, server, url, sockpath="/var/run/minissdpd.sock"):
    """ submits the specified service to MiniSSDPD (if running)"""
    # First check if sockpath exists i.e. MiniSSDPD is running
    if not os.path.exists(sockpath):
        return -1, f"Error: {sockpath} does not exist. Is minissdpd running?"
    # OK, submit
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        sock.connect(sockpath)
        sock.send(b'\x04' + codelength(st) + codelength(usn) + codelength(server) + codelength(url))
    except socket.error as msg:
        print(msg)
        return -1, msg
    finally:
        sock.close()
    return 0, "OK"


if __name__ == "__main__":
    # Example usage
    rc, message = submit_to_minissdpd(
        b'urn:schemas-upnp-org:device:InternetGatewayDevice:1',
        b'uuid:73616d61-6a6b-7a74-650a-0d24d4a5d636::urn:schemas-upnp-org:device:InternetGatewayDevice:1',
        b'MyServer/0.0',
        b'http://192.168.0.1:1234/rootDesc.xml',
    )
    if rc == 0:
        print("OK: submitting to MiniSSDPD went well")
    else:
        print("Not OK. Error message is:", message)



#***************************************************************************
#                                  _   _ ____  _
#  Project                     ___| | | |  _ \| |
#                             / __| | | | |_) | |
#                            | (__| |_| |  _ <| |___
#                             \___|\___/|_| \_\_____|
#
# Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://curl.se/docs/copyright.html.
#
# You may opt to use, copy, modify, merge, publish, distribute and/or sell
# copies of the Software, and permit persons to whom the Software is
# furnished to do so, under the terms of the COPYING file.
#
# This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
# KIND, either express or implied.
#
# SPDX-License-Identifier: curl
#
###########################################################################
#
import logging
import os
import sys
import platform
from typing import Generator, Union

import pytest

from testenv.env import EnvConfig

sys.path.append(os.path.join(os.path.dirname(__file__), '.'))

from testenv import Env, Nghttpx, Httpd, NghttpxQuic, NghttpxFwd

log = logging.getLogger(__name__)


def pytest_report_header(config):
    # Env inits its base properties only once, we can report them here
    env = Env()
    report = [
        f'Testing curl {env.curl_version()}',
        f'  platform: {platform.platform()}',
        f'  curl: Version: {env.curl_version_string()}',
        f'  curl: Features: {env.curl_features_string()}',
        f'  curl: Protocols: {env.curl_protocols_string()}',
        f'  httpd: {env.httpd_version()}',
        f'  httpd-proxy: {env.httpd_version()}'
    ]
    if env.have_h3():
        report.extend([
            f'  nghttpx: {env.nghttpx_version()}'
        ])
    if env.has_caddy():
        report.extend([
            f'  Caddy: {env.caddy_version()}'
        ])
    if env.has_vsftpd():
        report.extend([
            f'  VsFTPD: {env.vsftpd_version()}'
        ])
    buildinfo_fn = os.path.join(env.build_dir, 'buildinfo.txt')
    if os.path.exists(buildinfo_fn):
        with open(buildinfo_fn, 'r') as file_in:
            for line in file_in:
                line = line.strip()
                if line and not line.startswith('#'):
                    report.extend([line])
    return '\n'.join(report)


@pytest.fixture(scope='session')
def env_config(pytestconfig, testrun_uid, worker_id) -> EnvConfig:
    env_config = EnvConfig(pytestconfig=pytestconfig,
                           testrun_uid=testrun_uid,
                           worker_id=worker_id)
    return env_config


@pytest.fixture(scope='session', autouse=True)
def env(pytestconfig, env_config) -> Env:
    env = Env(pytestconfig=pytestconfig, env_config=env_config)
    level = logging.DEBUG if env.verbose > 0 else logging.INFO
    logging.getLogger('').setLevel(level=level)
    if not env.curl_has_protocol('http'):
        pytest.skip("curl built without HTTP support")
    if not env.curl_has_protocol('https'):
        pytest.skip("curl built without HTTPS support")
    if env.setup_incomplete():
        pytest.skip(env.incomplete_reason())

    env.setup()
    return env


@pytest.fixture(scope='session')
def httpd(env) -> Generator[Httpd, None, None]:
    httpd = Httpd(env=env)
    if not httpd.exists():
        pytest.skip(f'httpd not found: {env.httpd}')
    httpd.clear_logs()
    assert httpd.initial_start()
    yield httpd
    httpd.stop()


@pytest.fixture(scope='session')
def nghttpx(env, httpd) -> Generator[Union[Nghttpx,bool], None, None]:
    nghttpx = NghttpxQuic(env=env)
    if nghttpx.exists():
        if not nghttpx.supports_h3() and env.have_h3_curl():
            log.warning('nghttpx does not support QUIC, but curl does')
        nghttpx.clear_logs()
        assert nghttpx.initial_start()
        yield nghttpx
        nghttpx.stop()
    else:
        yield False


@pytest.fixture(scope='session')
def nghttpx_fwd(env, httpd) -> Generator[Union[Nghttpx,bool], None, None]:
    nghttpx = NghttpxFwd(env=env)
    if nghttpx.exists():
        nghttpx.clear_logs()
        assert nghttpx.initial_start()
        yield nghttpx
        nghttpx.stop()
    else:
        yield False


@pytest.fixture(scope='session')
def configures_httpd(env, httpd) -> Generator[bool, None, None]:
    # include this fixture as test parameter if the test configures httpd itself
    yield True

@pytest.fixture(scope='session')
def configures_nghttpx(env, httpd) -> Generator[bool, None, None]:
    # include this fixture as test parameter if the test configures nghttpx itself
    yield True

@pytest.fixture(autouse=True, scope='function')
def server_reset(request, env, httpd, nghttpx):
    # make sure httpd is in default configuration when a test starts
    if 'configures_httpd' not in request.node._fixtureinfo.argnames:
        httpd.reset_config()
        httpd.reload_if_config_changed()
    if env.have_h3() and \
            'nghttpx' in request.node._fixtureinfo.argnames and \
            'configures_nghttpx' not in request.node._fixtureinfo.argnames:
        nghttpx.reset_config()
        nghttpx.reload_if_config_changed()



from __future__ import print_function
import sys
import os
import timeit
import math
import argparse
import fnmatch
import json

parser = argparse.ArgumentParser(description="Python protobuf benchmark")
parser.add_argument("data_files", metavar="dataFile", nargs="+",
                    help="testing data files.")
parser.add_argument("--json", action="store_const", dest="json",
                    const="yes", default="no",
                    help="Whether to output json results")
parser.add_argument("--behavior_prefix", dest="behavior_prefix",
                    help="The output json format's behavior's name's prefix",
                    default="")
# BEGIN CPP GENERATED MESSAGE
parser.add_argument("--cpp_generated", action="store_const",
                    dest="cpp_generated", const="yes", default="no",
                    help="Whether to link generated code library")
# END CPP GENERATED MESSAGE
args = parser.parse_args()
# BEGIN CPP GENERATED MESSAGE
# CPP generated code must be linked before importing the generated Python code
# for the descriptor can be found in the pool
if args.cpp_generated != "no":
  sys.path.append( os.path.dirname( os.path.dirname( os.path.abspath(__file__) ) ) + "/.libs" )
  import libbenchmark_messages
  sys.path.append( os.path.dirname( os.path.dirname( os.path.abspath(__file__) ) ) + "/tmp" )
# END CPP GENERATED MESSAGE


import datasets.google_message1.proto2.benchmark_message1_proto2_pb2 as benchmark_message1_proto2_pb2
import datasets.google_message1.proto3.benchmark_message1_proto3_pb2 as benchmark_message1_proto3_pb2
import datasets.google_message2.benchmark_message2_pb2 as benchmark_message2_pb2
import datasets.google_message3.benchmark_message3_pb2 as benchmark_message3_pb2
import datasets.google_message4.benchmark_message4_pb2 as benchmark_message4_pb2
import benchmarks_pb2 as benchmarks_pb2


def run_one_test(filename):
  data = open(filename, "rb").read()
  benchmark_dataset = benchmarks_pb2.BenchmarkDataset()
  benchmark_dataset.ParseFromString(data)
  total_bytes = 0
  for payload in benchmark_dataset.payload:
    total_bytes += len(payload)
  benchmark_util = Benchmark(full_iteration=len(benchmark_dataset.payload),
                             module="py_benchmark",
                             setup_method="init",
                             total_bytes=total_bytes)
  result={}
  result["filename"] =  filename
  result["message_name"] =  benchmark_dataset.message_name
  result["benchmarks"] = {}
  benchmark_util.set_test_method("parse_from_benchmark")
  result["benchmarks"][args.behavior_prefix + "_parse_from_benchmark"] = \
    benchmark_util.run_benchmark(setup_method_args='"%s"' % (filename))
  benchmark_util.set_test_method("serialize_to_benchmark")
  result["benchmarks"][args.behavior_prefix + "_serialize_to_benchmark"] = \
    benchmark_util.run_benchmark(setup_method_args='"%s"' % (filename))
  return result


def init(filename):
  global benchmark_dataset, message_class, message_list, counter, total_bytes
  message_list=[]
  counter = 0
  total_bytes = 0
  data = open(filename, "rb").read()
  benchmark_dataset = benchmarks_pb2.BenchmarkDataset()
  benchmark_dataset.ParseFromString(data)

  if benchmark_dataset.message_name == "benchmarks.proto3.GoogleMessage1":
    message_class = benchmark_message1_proto3_pb2.GoogleMessage1
  elif benchmark_dataset.message_name == "benchmarks.proto2.GoogleMessage1":
    message_class = benchmark_message1_proto2_pb2.GoogleMessage1
  elif benchmark_dataset.message_name == "benchmarks.proto2.GoogleMessage2":
    message_class = benchmark_message2_pb2.GoogleMessage2
  elif benchmark_dataset.message_name == "benchmarks.google_message3.GoogleMessage3":
    message_class = benchmark_message3_pb2.GoogleMessage3
  elif benchmark_dataset.message_name == "benchmarks.google_message4.GoogleMessage4":
    message_class = benchmark_message4_pb2.GoogleMessage4
  else:
    raise IOError("Message %s not found!" % (benchmark_dataset.message_name))

  for one_payload in benchmark_dataset.payload:
    temp = message_class()
    temp.ParseFromString(one_payload)
    message_list.append(temp)
    total_bytes += len(one_payload)


def parse_from_benchmark():
  global counter, message_class, benchmark_dataset
  m = message_class().ParseFromString(benchmark_dataset.payload[counter % len(benchmark_dataset.payload)])
  counter = counter + 1


def serialize_to_benchmark():
  global counter, message_list, message_class
  s = message_list[counter % len(benchmark_dataset.payload)].SerializeToString()
  counter = counter + 1


class Benchmark:
  def __init__(self, module=None, test_method=None,
               setup_method=None, total_bytes=None, full_iteration = 1):
    self.full_iteration = full_iteration
    self.module = module
    self.test_method = test_method
    self.setup_method = setup_method
    self.total_bytes = total_bytes

  def set_test_method(self, test_method):
    self.test_method = test_method

  def full_setup_code(self, setup_method_args=''):
    setup_code = ""
    setup_code += "from %s import %s\n" % (self.module, self.test_method)
    setup_code += "from %s import %s\n" % (self.module, self.setup_method)
    setup_code += "%s(%s)\n" % (self.setup_method, setup_method_args)
    return setup_code

  def dry_run(self, test_method_args='', setup_method_args=''):
    return timeit.timeit(stmt="%s(%s)" % (self.test_method, test_method_args),
                         setup=self.full_setup_code(setup_method_args),
                         number=self.full_iteration);

  def run_benchmark(self, test_method_args='', setup_method_args=''):
    reps = self.full_iteration;
    t = self.dry_run(test_method_args, setup_method_args);
    if t < 3 :
      reps = int(math.ceil(3 / t)) * self.full_iteration
    if reps != self.full_iteration:
        t = timeit.timeit(stmt="%s(%s)" % (self.test_method, test_method_args),
                          setup=self.full_setup_code(setup_method_args),
                          number=reps);
    return self.total_bytes * 1.0 / 2 ** 20 / (1.0 * t / reps * self.full_iteration)


if __name__ == "__main__":
  results = []
  for file in args.data_files:
    results.append(run_one_test(file))

  if args.json != "no":
    print(json.dumps(results))
  else:
    for result in results:
      print("Message %s of dataset file %s" % \
          (result["message_name"], result["filename"]))
      print("Average throughput for parse_from_benchmark: %.2f MB/s" % \
          (result["benchmarks"][ \
                      args.behavior_prefix + "_parse_from_benchmark"]))
      print("Average throughput for serialize_to_benchmark: %.2f MB/s" % \
          (result["benchmarks"][ \
                      args.behavior_prefix + "_serialize_to_benchmark"]))
      print("")



#!/usr/bin/env python
"""
compare_bench.py - Compare two benchmarks or their results and report the
                   difference.
"""
import argparse
from argparse import ArgumentParser
import sys
import gbench
from gbench import util, report
from gbench.util import *

def check_inputs(in1, in2, flags):
    """
    Perform checking on the user provided inputs and diagnose any abnormalities
    """
    in1_kind, in1_err = classify_input_file(in1)
    in2_kind, in2_err = classify_input_file(in2)
    output_file = find_benchmark_flag('--benchmark_out=', flags)
    output_type = find_benchmark_flag('--benchmark_out_format=', flags)
    if in1_kind == IT_Executable and in2_kind == IT_Executable and output_file:
        print(("WARNING: '--benchmark_out=%s' will be passed to both "
              "benchmarks causing it to be overwritten") % output_file)
    if in1_kind == IT_JSON and in2_kind == IT_JSON and len(flags) > 0:
        print("WARNING: passing --benchmark flags has no effect since both "
              "inputs are JSON")
    if output_type is not None and output_type != 'json':
        print(("ERROR: passing '--benchmark_out_format=%s' to 'compare_bench.py`"
              " is not supported.") % output_type)
        sys.exit(1)


def main():
    parser = ArgumentParser(
        description='compare the results of two benchmarks')
    parser.add_argument(
        'test1', metavar='test1', type=str, nargs=1,
        help='A benchmark executable or JSON output file')
    parser.add_argument(
        'test2', metavar='test2', type=str, nargs=1,
        help='A benchmark executable or JSON output file')
    # FIXME this is a dummy argument which will never actually match
    # any --benchmark flags but it helps generate a better usage message
    parser.add_argument(
        'benchmark_options', metavar='benchmark_option', nargs='*',
        help='Arguments to pass when running benchmark executables'
    )
    args, unknown_args = parser.parse_known_args()
    # Parse the command line flags
    test1 = args.test1[0]
    test2 = args.test2[0]
    if args.benchmark_options:
        print("Unrecognized positional argument arguments: '%s'"
              % args.benchmark_options)
        exit(1)
    benchmark_options = unknown_args
    check_inputs(test1, test2, benchmark_options)
    # Run the benchmarks and report the results
    json1 = gbench.util.run_or_load_benchmark(test1, benchmark_options)
    json2 = gbench.util.run_or_load_benchmark(test2, benchmark_options)
    output_lines = gbench.report.generate_difference_report(json1, json2)
    print('Comparing %s to %s' % (test1, test2))
    for ln in output_lines:
        print(ln)


if __name__ == '__main__':
    main()



#
# dotex: "I don't care, just do the tex or whatever..."
#
# Written in 2025 by Lexi Mayfield
# Released into the Public Domain, see UNLICENSE.txt
#

import io
import logging
import struct
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional


@dataclass
class Lump:
    filepos: int
    size: int
    name: str


class DoomWADWriter:
    name: str
    handle: io.BufferedRandom
    lumps: List[Lump]

    def __init__(self, file: Path):
        self.name = str(file.name)
        self.lumps = []
        self.handle = open(file, "wb+")
        self.handle.write(b"\x00" * 12)  # Blank WAD header

    def write_lump(self, name: str, data: Optional[bytes] = None) -> None:
        """Write out lump data, while also noting its name and location."""
        logging.debug(f"Writing lump '{name}' to '{self.name}'.")

        if data is None:
            self.lumps.append(Lump(filepos=0, size=0, name=name))
            return

        self.lumps.append(Lump(filepos=self.handle.tell(), size=len(data), name=name))
        self.handle.write(data)

    def finalize(self, iwad: bool = False):
        """
        Finalize the WAD file - write out the directory, fix up the header
        to match, then close the file.
        """
        logging.debug(f"Finalizing WAD '{self.name}'.")

        # Pad to 4 bytes.
        while self.handle.tell() % 4 != 0:
            self.handle.write(b"\x00")

        infotableofs = self.handle.tell()

        for lump in self.lumps:
            self.handle.write(
                struct.pack("<II8s", lump.filepos, lump.size, lump.name.encode("ascii"))
            )

        self.handle.seek(0)
        if iwad:
            self.handle.write(b"IWAD")
        else:
            self.handle.write(b"PWAD")

        self.handle.write(struct.pack("<II", len(self.lumps), infotableofs))
        self.handle.close()



#! /usr/bin/env python
# encoding: utf-8

import argparse
import errno
import logging
import os
import platform
import re
import sys
import subprocess
import tempfile

try:
    import winreg
except ImportError:
    import _winreg as winreg
try:
    import urllib.request as request
except ImportError:
    import urllib as request
try:
    import urllib.parse as parse
except ImportError:
    import urlparse as parse

class EmptyLogger(object):
    '''
    Provides an implementation that performs no logging
    '''
    def debug(self, *k, **kw):
        pass
    def info(self, *k, **kw):
        pass
    def warn(self, *k, **kw):
        pass
    def error(self, *k, **kw):
        pass
    def critical(self, *k, **kw):
        pass
    def setLevel(self, *k, **kw):
        pass

urls = (
    'http://downloads.sourceforge.net/project/mingw-w64/Toolchains%20'
        'targetting%20Win32/Personal%20Builds/mingw-builds/installer/'
        'repository.txt',
    'http://downloads.sourceforge.net/project/mingwbuilds/host-windows/'
        'repository.txt'
)
'''
A list of mingw-build repositories
'''

def repository(urls = urls, log = EmptyLogger()):
    '''
    Downloads and parse mingw-build repository files and parses them
    '''
    log.info('getting mingw-builds repository')
    versions = {}
    re_sourceforge = re.compile(r'http://sourceforge.net/projects/([^/]+)/files')
    re_sub = r'http://downloads.sourceforge.net/project/\1'
    for url in urls:
        log.debug(' - requesting: %s', url)
        socket = request.urlopen(url)
        repo = socket.read()
        if not isinstance(repo, str):
            repo = repo.decode();
        socket.close()
        for entry in repo.split('\n')[:-1]:
            value = entry.split('|')
            version = tuple([int(n) for n in value[0].strip().split('.')])
            version = versions.setdefault(version, {})
            arch = value[1].strip()
            if arch == 'x32':
                arch = 'i686'
            elif arch == 'x64':
                arch = 'x86_64'
            arch = version.setdefault(arch, {})
            threading = arch.setdefault(value[2].strip(), {})
            exceptions = threading.setdefault(value[3].strip(), {})
            revision = exceptions.setdefault(int(value[4].strip()[3:]),
                re_sourceforge.sub(re_sub, value[5].strip()))
    return versions

def find_in_path(file, path=None):
    '''
    Attempts to find an executable in the path
    '''
    if platform.system() == 'Windows':
        file += '.exe'
    if path is None:
        path = os.environ.get('PATH', '')
    if type(path) is type(''):
        path = path.split(os.pathsep)
    return list(filter(os.path.exists,
        map(lambda dir, file=file: os.path.join(dir, file), path)))

def find_7zip(log = EmptyLogger()):
    '''
    Attempts to find 7zip for unpacking the mingw-build archives
    '''
    log.info('finding 7zip')
    path = find_in_path('7z')
    if not path:
        key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r'SOFTWARE\7-Zip')
        path, _ = winreg.QueryValueEx(key, 'Path')
        path = [os.path.join(path, '7z.exe')]
    log.debug('found \'%s\'', path[0])
    return path[0]

find_7zip()

def unpack(archive, location, log = EmptyLogger()):
    '''
    Unpacks a mingw-builds archive
    '''
    sevenzip = find_7zip(log)
    log.info('unpacking %s', os.path.basename(archive))
    cmd = [sevenzip, 'x', archive, '-o' + location, '-y']
    log.debug(' - %r', cmd)
    with open(os.devnull, 'w') as devnull:
        subprocess.check_call(cmd, stdout = devnull)

def download(url, location, log = EmptyLogger()):
    '''
    Downloads and unpacks a mingw-builds archive
    '''
    log.info('downloading MinGW')
    log.debug(' - url: %s', url)
    log.debug(' - location: %s', location)

    re_content = re.compile(r'attachment;[ \t]*filename=(")?([^"]*)(")?[\r\n]*')

    stream = request.urlopen(url)
    try:
        content = stream.getheader('Content-Disposition') or ''
    except AttributeError:
        content = stream.headers.getheader('Content-Disposition') or ''
    matches = re_content.match(content)
    if matches:
        filename = matches.group(2)
    else:
        parsed = parse.urlparse(stream.geturl())
        filename = os.path.basename(parsed.path)

    try:
        os.makedirs(location)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(location):
            pass
        else:
            raise

    archive = os.path.join(location, filename)
    with open(archive, 'wb') as out:
        while True:
            buf = stream.read(1024)
            if not buf:
                break
            out.write(buf)
    unpack(archive, location, log = log)
    os.remove(archive)

    possible = os.path.join(location, 'mingw64')
    if not os.path.exists(possible):
        possible = os.path.join(location, 'mingw32')
        if not os.path.exists(possible):
            raise ValueError('Failed to find unpacked MinGW: ' + possible)
    return possible

def root(location = None, arch = None, version = None, threading = None,
        exceptions = None, revision = None, log = EmptyLogger()):
    '''
    Returns the root folder of a specific version of the mingw-builds variant
    of gcc. Will download the compiler if needed
    '''

    # Get the repository if we don't have all the information
    if not (arch and version and threading and exceptions and revision):
        versions = repository(log = log)

    # Determine some defaults
    version = version or max(versions.keys())
    if not arch:
        arch = platform.machine().lower()
        if arch == 'x86':
            arch = 'i686'
        elif arch == 'amd64':
            arch = 'x86_64'
    if not threading:
        keys = versions[version][arch].keys()
        if 'posix' in keys:
            threading = 'posix'
        elif 'win32' in keys:
            threading = 'win32'
        else:
            threading = keys[0]
    if not exceptions:
        keys = versions[version][arch][threading].keys()
        if 'seh' in keys:
            exceptions = 'seh'
        elif 'sjlj' in keys:
            exceptions = 'sjlj'
        else:
            exceptions = keys[0]
    if revision == None:
        revision = max(versions[version][arch][threading][exceptions].keys())
    if not location:
        location = os.path.join(tempfile.gettempdir(), 'mingw-builds')

    # Get the download url
    url = versions[version][arch][threading][exceptions][revision]

    # Tell the user whatzzup
    log.info('finding MinGW %s', '.'.join(str(v) for v in version))
    log.debug(' - arch: %s', arch)
    log.debug(' - threading: %s', threading)
    log.debug(' - exceptions: %s', exceptions)
    log.debug(' - revision: %s', revision)
    log.debug(' - url: %s', url)

    # Store each specific revision differently
    slug = '{version}-{arch}-{threading}-{exceptions}-rev{revision}'
    slug = slug.format(
        version = '.'.join(str(v) for v in version),
        arch = arch,
        threading = threading,
        exceptions = exceptions,
        revision = revision
    )
    if arch == 'x86_64':
        root_dir = os.path.join(location, slug, 'mingw64')
    elif arch == 'i686':
        root_dir = os.path.join(location, slug, 'mingw32')
    else:
        raise ValueError('Unknown MinGW arch: ' + arch)

    # Download if needed
    if not os.path.exists(root_dir):
        downloaded = download(url, os.path.join(location, slug), log = log)
        if downloaded != root_dir:
            raise ValueError('The location of mingw did not match\n%s\n%s'
                % (downloaded, root_dir))

    return root_dir

def str2ver(string):
    '''
    Converts a version string into a tuple
    '''
    try:
        version = tuple(int(v) for v in string.split('.'))
        if len(version) is not 3:
            raise ValueError()
    except ValueError:
        raise argparse.ArgumentTypeError(
            'please provide a three digit version string')
    return version

def main():
    '''
    Invoked when the script is run directly by the python interpreter
    '''
    parser = argparse.ArgumentParser(
        description = 'Downloads a specific version of MinGW',
        formatter_class = argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('--location',
        help = 'the location to download the compiler to',
        default = os.path.join(tempfile.gettempdir(), 'mingw-builds'))
    parser.add_argument('--arch', required = True, choices = ['i686', 'x86_64'],
        help = 'the target MinGW architecture string')
    parser.add_argument('--version', type = str2ver,
        help = 'the version of GCC to download')
    parser.add_argument('--threading', choices = ['posix', 'win32'],
        help = 'the threading type of the compiler')
    parser.add_argument('--exceptions', choices = ['sjlj', 'seh', 'dwarf'],
        help = 'the method to throw exceptions')
    parser.add_argument('--revision', type=int,
        help = 'the revision of the MinGW release')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-v', '--verbose', action='store_true',
        help='increase the script output verbosity')
    group.add_argument('-q', '--quiet', action='store_true',
        help='only print errors and warning')
    args = parser.parse_args()

    # Create the logger
    logger = logging.getLogger('mingw')
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    if args.quiet:
        logger.setLevel(logging.WARN)
    if args.verbose:
        logger.setLevel(logging.DEBUG)

    # Get MinGW
    root_dir = root(location = args.location, arch = args.arch,
        version = args.version, threading = args.threading,
        exceptions = args.exceptions, revision = args.revision,
        log = logger)

    sys.stdout.write('%s\n' % os.path.join(root_dir, 'bin'))

if __name__ == '__main__':
    try:
        main()
    except IOError as e:
        sys.stderr.write('IO error: %s\n' % e)
        sys.exit(1)
    except OSError as e:
        sys.stderr.write('OS error: %s\n' % e)
        sys.exit(1)
    except KeyboardInterrupt as e:
        sys.stderr.write('Killed\n')
        sys.exit(1)



#! /usr/bin/env python
# vim: tabstop=2 shiftwidth=2 expandtab
# MiniUPnP project
# Author : Thomas Bernard
# Python 3
# This Sample code is public domain.
# website : https://miniupnp.tuxfamily.org/

# import the python miniupnpc module
import miniupnpc
import sys

try:
  import argparse
  parser = argparse.ArgumentParser()
  parser.add_argument('-m', '--multicastif')
  parser.add_argument('-p', '--minissdpdsocket')
  parser.add_argument('-d', '--discoverdelay', type=int, default=200)
  parser.add_argument('-z', '--localport', type=int, default=0)
  # create the object
  u = miniupnpc.UPnP(**vars(parser.parse_args()))
except:
  print('argparse not available')
  i = 1
  multicastif = None
  minissdpdsocket = None
  discoverdelay = 200
  localport = 0
  while i < len(sys.argv):
    print(sys.argv[i])
    if sys.argv[i] == '-m' or sys.argv[i] == '--multicastif':
      multicastif = sys.argv[i+1]
    elif sys.argv[i] == '-p' or sys.argv[i] == '--minissdpdsocket':
      minissdpdsocket = sys.argv[i+1]
    elif sys.argv[i] == '-d' or sys.argv[i] == '--discoverdelay':
      discoverdelay = int(sys.argv[i+1])
    elif sys.argv[i] == '-z' or sys.argv[i] == '--localport':
      localport = int(sys.argv[i+1])
    else:
      raise Exception('invalid argument %s' % sys.argv[i])
    i += 2
  # create the object
  u = miniupnpc.UPnP(multicastif, minissdpdsocket, discoverdelay, localport)

print('inital(default) values :')
print(' discoverdelay', u.discoverdelay)
print(' lanaddr', u.lanaddr)
print(' multicastif', u.multicastif)
print(' minissdpdsocket', u.minissdpdsocket)
#u.minissdpdsocket = '../minissdpd/minissdpd.sock'
# discovery process, it usually takes several seconds (2 seconds or more)
print('Discovering... delay=%ums' % u.discoverdelay)
print('%d device(s) detected' % u.discover())
# select an igd
try:
  u.selectigd()
except Exception as e:
  print('Exception :', e)
  sys.exit(1)
# it is also possible to pass the root description URL to u.selectigd() :
# u.selectigd('http://192.168.1.254:5678/desc/root')
# display information about the IGD and the internet connection
print('local ip address :', u.lanaddr)
print('external ip address :', u.externalipaddress())
print( u.statusinfo(), u.connectiontype())
print('total bytes : sent', u.totalbytesent(), 'received', u.totalbytereceived())
print('total packets : sent', u.totalpacketsent(), 'received', u.totalpacketreceived())

#print u.addportmapping(64000, 'TCP',
#                       '192.168.1.166', 63000, 'port mapping test', '')
#print u.deleteportmapping(64000, 'TCP')

port = 0
proto = 'UDP'
# list the redirections :
i = 0
while True:
	p = u.getgenericportmapping(i)
	if p==None:
		break
	print(i, p)
	(port, proto, (ihost,iport), desc, c, d, e) = p
	#print port, desc
	i = i + 1

print(u.getspecificportmapping(port, proto))
try:
  print(u.getportmappingnumberofentries())
except Exception as e:
  print('GetPortMappingNumberOfEntries() is not supported :', e)




#!/usr/bin/python3

import sys
import time
import signal
import getopt
import pyaudio
import platform

from ctypes import *

if platform.system() == 'Windows':
    adl = CDLL("libADLMIDI.dll")
else:
    adl = CDLL("libADLMIDI.so")

_FREQUENCY = 44100

playing = True

# Allows to quit by Ctrl+C
def catch_signal(signum, frame):
    global playing
    playing = False
    print("Caught Ctrl+C!")


class MidiPlayer:
    """A streaming IMF music player."""
    SAMPLE_SIZE = 2  # 16-bit
    CHANNELS = 2  # stereo
    BUFFER_SAMPLE_COUNT = 1024  # Max allowed: 512

    def __init__(self, freq=_FREQUENCY):
        """Initializes the PyAudio and OPL synth."""
        self._freq = freq
        # Prepare PyAudio
        self._audio = pyaudio.PyAudio()
        # Prepare buffers.
        self._data = create_string_buffer(MidiPlayer.BUFFER_SAMPLE_COUNT * MidiPlayer.SAMPLE_SIZE * MidiPlayer.CHANNELS)
        self._stream = None  # Created later.
        self._opl = adl.adl_init(freq)
        self._song = None

    def set_bank_no(self, bank_no):
        if adl.adl_setBank(self._opl, bank_no):
            print("%s" % c_char_p(adl.adl_errorInfo(self._opl)).value)
            return False
        return True

    def set_bank_file(self, bank_file):
        if adl.adl_openBankFile(self._opl, bank_file.encode()) < 0:
            print("%s" % c_char_p(adl.adl_errorInfo(self._opl)).value)
            return False
        return True

    def set_loop_enabled(self, loop_en):
        adl.adl_setLoopEnabled(self._opl, 1 if loop_en else 0)

    def _create_stream(self, start=True):
        """Create a new PyAudio stream."""
        self._stream = self._audio.open(
            format=self._audio.get_format_from_width(MidiPlayer.SAMPLE_SIZE),
            channels=MidiPlayer.CHANNELS,
            frames_per_buffer=MidiPlayer.BUFFER_SAMPLE_COUNT,
            rate=self._freq,
            output=True,
            start=start,  # Don't start playing immediately!
            stream_callback=self._callback)

    def set_song(self, song):
        self._song = song
        if adl.adl_openFile(self._opl, self._song.encode()) < 0:
            print("%s" % c_char_p(adl.adl_errorInfo(self._opl)).value.decode())
            return False

        return True

    # noinspection PyUnusedLocal
    def _callback(self, input_data, frame_count, time_info, status):
        global playing
        got = adl.adl_play(self._opl, int(MidiPlayer.BUFFER_SAMPLE_COUNT * 2), cast(self._data, POINTER(c_short)));
        if got <= 0:
            playing = False
            return None, pyaudio.paComplete
        return self._data, pyaudio.paContinue

    def play(self, repeat=False):
        global playing
        """Starts playing the song at the current position."""
        self.repeat = repeat
        if self._song is None:
            playing = False
            return
        # If a stream exists and it is not active and not stopped, it needs to be closed and a new one created.
        if self._stream is not None and not self._stream.is_active() and not self._stream.is_stopped():
            self._stream.close()  # close for good measure.
            self._stream = None

        # If there's no stream at this point, create one.
        if self._stream is None:
            self._create_stream(False)
        self._stream.start_stream()

    def pause(self):
        """Stops the PyAudio stream, but does not rewind the playback position."""
        self._stream.stop_stream()

    def stop(self):
        """Stops the PyAudio stream and resets the playback position."""
        self._stream.stop_stream()

    def close(self):
        """Closes the PyAudio stream and terminates it."""
        if self._stream is not None:
            self._stream.stop_stream()
            self._stream.close()
            self._stream = None
        self._audio.terminate()
        adl.adl_close(self._opl)


def usage(arg):
    print('usage: %s [-b <embedded bank number>|-bf <external bank file>] [-l 0|1] <file>' % arg, file=sys.stderr)
    sys.exit(2)

if __name__ == '__main__':
    bank_no = 72
    bank_file = None
    enable_loop = False
    sample_rate = 44100

    opts, args = getopt.getopt(sys.argv[1:], 'b:bg:l:')
    for o, a in opts:
        if o == '-b':
            bank_no = int(a)
        if o == '-l':
            enable_loop = int(a) == 1
        elif o == '-bg':
            bank_file = a

    if not args:
        usage(sys.argv[0])

    midi_file = args[0]

    MIDIPlay = MidiPlayer(sample_rate)

    if bank_file:
        if not MIDIPlay.set_bank_file(bank_file):
            exit(1)
        print("Using bank file: %s" % bank_file)
    else:
        if not MIDIPlay.set_bank_no(bank_no):
            exit(1)
        print("Using embedded bank: %d" % bank_no)

    MIDIPlay.set_loop_enabled(enable_loop)

    if not MIDIPlay.set_song(midi_file):
        exit(1)

    MIDIPlay.play()

    signal.signal(signal.SIGINT, catch_signal)

    print("Playing MIDI file %s..." % (midi_file))
    while playing:
        time.sleep(0.1)
    MIDIPlay.stop()

    MIDIPlay.close()




#!/usr/bin/python3

import sys
import signal
import wave
import getopt
import alsaaudio

from ctypes import *

# adl = CDLL("./libADLMIDI.so")
adl = CDLL("/usr/local/lib/libADLMIDI.so")

#class ADLMIDI_AudioFormat(Structure):
#     _fields_ = [("type", c_int),
#                 ("containerSize", c_uint),
#                 ("outasampleOffset", c_uint)]

playing = True

# Allows to quit by Ctrl+C
def catch_signal(signum, frame):
    global playing
    playing = False
    print("Caught Ctrl+C!")

def play(device, MIDIPlay, sampleRate):
    global playing

    # Opening ALSA output device
    device = alsaaudio.PCM(device=device)

    # Set attributes
    device.setchannels(2)
    device.setrate(sampleRate)
    device.setperiodsize(512)

    # format = ADLMIDI_AudioFormat(0, 2, 4);
    device.setformat(alsaaudio.PCM_FORMAT_S16_LE)

    # Preparing the audio output buffer
    data = create_string_buffer(4096 * 2);
    dataLen = 4096

    print("Streaming...")

    # Streaming the audio data into ALSA output
    while playing:
        got = adl.adl_play(MIDIPlay, dataLen, cast(data, POINTER(c_short)));
        if got <= 0:
            break;
        device.write(bytearray(data.raw))

    # Closing the audio output device
    device.close()

    print("Quiting...")


def usage():
    print('usage: playmidi.py <file>', file=sys.stderr)
    sys.exit(2)

if __name__ == '__main__':
    device = 'default'
    sampleRate = 44100

    opts, args = getopt.getopt(sys.argv[1:], 'd:')
    for o, a in opts:
        if o == '-d':
            device = a

    if not args:
        usage()

    midiFile = args[0]

    # Initialize the library
    MIDIPlay = adl.adl_init(sampleRate)

    # ============= Optional setup ====================
    # Turn on looping (optionally)
    # adl.adl_setLoopEnabled(MIDIPlay, 1)

    # Changing the bank number (optionally)
    # if adl.adl_setBank(MIDIPlay, 68) < 0:
    #    print(c_char_p(adl.adl_errorInfo(MIDIPlay)).value)
    #    sys.exit(1)
    # ============= Optional setup =End================

    # Open music file...
    if adl.adl_openFile(MIDIPlay, midiFile.encode()) < 0:
        print(c_char_p(adl.adl_errorInfo(MIDIPlay)).value)
        sys.exit(1)

    print("Playing MIDI file %s..." % (midiFile))

    # Allows to quit by Ctrl+C
    signal.signal(signal.SIGINT, catch_signal)

    # Start the audio streaming
    play(device, MIDIPlay, sampleRate)

    # Closing the library
    adl.adl_close(MIDIPlay)

